
#include "command_mode.h"

extern bool QUIT_VITUNES;

/*****************************************************************************
 *
 *                       Command-Window Commands
 *
 * What follows are the functions handlers for each of the commands that can
 * be executed in the command-window (things like ":q" for quit, ":w" for
 * saving, etc.  Each function name is preceeded with "cwcmd_" and each
 * takes the form of argc+argv[] argument list, and user interface, media
 * library, and mediaplayer.
 *
 * Note that for all of them, a proper error-free run should return 0,
 * otherwise something nonzero should be returned and the global variable
 * vitunes_error (char*) should be set to something appropriate.
 *
 * Care should be taken in any functions called to make sure they don't
 * error/log to stdout/stderr.
 *
 ****************************************************************************/

#define VITUNES_ERR_SIZE 1000
char vitunes_error[VITUNES_ERR_SIZE];

/* command to quit vitunes */
int cwcmd_quit(int argc, char *argv[], vitunes *v)
{
   if (argc != 1)
   {  snprintf(vitunes_error, VITUNES_ERR_SIZE, "usage :q");
      return 1;
   }

   QUIT_VITUNES = true;
   return 0;
}

/* command to save the currently viewed playlist */
int cwcmd_write(int argc, char *argv[], vitunes *v)
{
   if (argc > 2)
   {  snprintf(vitunes_error, VITUNES_ERR_SIZE, "usage :w[!] [name]");
      return 1;
   }

   /* is this write forced? */
   bool forced = false;
   if (strcmp(argv[0], "w!") == 0)
      forced = true;

   /* get the current playlist (the one we're writing */
   int pindex = v->viewing_playlist;
   playlist *plist = v->library->playlists[pindex];

   /* saving current playlist */
   if (argc == 1)
   {  /* can't save current playlist... */
      if (pindex == ML_PINDEX_LIBRARY || pindex == ML_PINDEX_SRESULTS)
      {  snprintf(vitunes_error, VITUNES_ERR_SIZE,
            "use \":w name\" for pseudo-playlists");
         return 1;
      }

      /* can't save new playlist without a name (e.g. need ":w name" form) */
      if (plist->filename == NULL)
      {  snprintf(vitunes_error, VITUNES_ERR_SIZE,
            "usage \":w name\" for new playlists");
         return 1;
      }

      playlist_save(plist);
      return 0;
   }

   /* "Save As..." form of saving current playlist, argv[1] is name */
   if (argc == 2)
   {
      static char filename[PATH_MAX + 1];

      /* build the filename */
      sprintf(filename, "%s/%s.playlist", v->library->playlist_dir, argv[1]);

      /* check if a playlist with that filename already exists */
      int new_pindex = -1;
      int i;
      for (i = 0; i < v->library->playlists_num; i++)
      {  if (v->library->playlists[i]->filename != NULL
         && strcmp(v->library->playlists[i]->filename, filename) == 0)
            new_pindex = i;
      }

      /* it's the same one, just save */
      if (pindex == new_pindex)
      {  playlist_save(plist);
         return 0;
      }

      /* if such a playlist already exists and we weren't forced, error */
      if (new_pindex > 0 && !forced)
      {  snprintf(vitunes_error, VITUNES_ERR_SIZE,
            "playlist \"%s\" already exists.  use \":w! %s\" to force",
            argv[1], argv[1]);
         return 1;
      }

      /* duplicate the playlist we're saving */
      playlist *new_plist = playlist_dup(plist, filename, argv[1]);

      /* if we're replacing one, remove it.  otherwise, figure out where it
       * will go in media library's playlist array */
      if (new_pindex > 0)
         ml_remove_playlist(v->library, new_pindex);
      else
         v->ui->playlists->max_rows++;


      /* save playlist and plug it into the media library */
      playlist_save(new_plist);
      ml_add_playlist(v->library, new_plist);
      viui_draw_playlists(v);
      return 0;
   }

   /* shouldn't reach here */
   return 1;
}

/* command to change the current play mode */
int cwcmd_mode(int argc, char *argv[], vitunes *v)
{
   if (argc != 2)
   {  snprintf(vitunes_error, VITUNES_ERR_SIZE,
         "usage: mode [ linear | loop | random ]");
      return 1;
   }

   if (strcmp(argv[1], "linear") == 0)
      v->mplayer->mode = MP_MODE_LINEAR;
   else if (strcmp(argv[1], "loop") == 0)
      v->mplayer->mode = MP_MODE_LOOP;
   else if (strcmp(argv[1], "random") == 0)
      v->mplayer->mode = MP_MODE_RANDOM;
   else
   {  snprintf(vitunes_error, VITUNES_ERR_SIZE,
         "usage: mode [ linear | loop | random ]");
      return 1;
   }

   return 0;
}

/* command to create new playlist */
int cwcmd_new(int argc, char *argv[], vitunes *v)
{
   if (argc <= 1)
   {  snprintf(vitunes_error, VITUNES_ERR_SIZE, "usage: new [ playlist ]");
      return 1;
   }

   if (strcmp(argv[1], "playlist") == 0)
   {  /* create new playlist */
      playlist *p = playlist_new();
      p->name = strdup("untitled");
      if (p->name == NULL)
         err(1, "unable to set name of new playlist");

      /* plug it into the media library */
      ml_add_playlist(v->library, p);
      v->ui->playlists->max_rows++;
      viui_draw_playlists(v);
      return 0;
   }

   snprintf(vitunes_error, VITUNES_ERR_SIZE, "usage: new [ playlist ]");
   return 1;
}

/* array of the above functions and their names */
struct cwcmd_entry {
   char  *name;
   int   (*fname)(int argc, char *argv[], vitunes *v);
};

const struct cwcmd_entry CWCMD_PATH[] = {
   { "q",      cwcmd_quit },
   { "w",      cwcmd_write },
   { "w!",     cwcmd_write },
   { "mode",   cwcmd_mode },
   { "new",    cwcmd_new }
};
const int CWCMD_PATH_SIZE = (sizeof(CWCMD_PATH) / sizeof(struct cwcmd_entry));

/* function to take a command and tokenize it into an argc/argv style */
void str2argv(const char *str, int *argc, char ***argv)
{
   /* TODO This should handle some rudimentary escape sequence the way
    * argc/argv are normally built.
    * For example, currently the command:
    *    :w foo\ bar
    * Will create an argc/argv pair of "3" and { ":w", "foo\", "bar" }
    * when it should, intuitively (to match normal semantics), create an
    * argc of "2" and argv of { ":w" , "foo bar" }.
    */

   char *token;
   char *line = strdup(str);
   if (line == NULL)
      err(1, "failed to duplicate string");

   *argv = (char**) calloc(500, sizeof(char*));
   if (*argv == NULL)
      err(1, "failed to allocate for argv");

   bzero(*argv, 500 * sizeof(char*));

   *argc = 0;
   while ((token = strsep(&line, " \t\n")) != NULL)
   {  if (strlen(token) == 0)
         continue;
      else
      {  (*argv)[ *argc ] = strdup(token);
         if ((*argv)[ *argc ] == NULL)
            err(1, "failed to strdup");

         *argc = *argc + 1;
      }
   }

   free(line);
}

void argvfree(int *argc, char ***argv)
{  int i;
   for (i = 0; i < *argc; i++)
      free((*argv)[i]);

   free(*argv);
   *argc = 0;
}


/*****************************************************************************
 * Command-Window Input Handler
 ****************************************************************************/
void command_mode_execute(const char *command, vitunes *v)
{
   int     argc;
   char  **argv;

   /* Parse the command into an argc/argv structure */
   str2argv(command, &argc, &argv);
   if (argc == 0)
   {  argvfree(&argc, &argv);
      viui_draw_row(v->ui->command, 0, UI_LEFT, "");
      wrefresh(v->ui->command->cwin);
      return;
   }

   /* Search the 'path' for a command of the name provided */
   bool found   = false;
   int  retCode = 0;
   int  i;
   for (i = 0; i < CWCMD_PATH_SIZE && !found; i++)
   {  if (strcmp(argv[0], CWCMD_PATH[i].name) == 0)
      {  found = true;
         retCode = (CWCMD_PATH[i].fname)(argc, argv, v);
      }
   }

   /* If no such command was found -or- there were errors executing the
    * command, display those errors now... */
   if (!found)
      viui_draw_command_error(v, "unknown command \"%s\"", argv[0]);
   else if (retCode != 0)
      viui_draw_command_error(v, vitunes_error);


   /* cleanup */
   argvfree(&argc, &argv);
   wrefresh(v->ui->command->cwin);
   return;
}


