
#include "input_handler.h"

/* Simple function to setup the copy buffer after a "yank" or "delete"
 * operation.
 */
void copy_buff_set(vitunes *v, int size)
{
   /* free anything currently in the copy buff */
   if (v->copy_buff_size > 0)
   {  free(v->copy_buff);
      v->copy_buff_size = 0;
   }

   /* setup new copy buff */
   v->copy_buff = calloc(size, sizeof(meta_info*));
   if (v->copy_buff == NULL)
      err(1, "failed to create copy buffer");

   /* get info about where we're copying from */
   playlist *p = v->library->playlists[v->viewing_playlist];
   int   start = v->ui->medialist->cur_row + v->ui->medialist->voffset;

   /* copy file refs to copy buff */
   int i;
   for (i = 0; i < size; i++)
      v->copy_buff[i] = p->files[start + i];

   v->copy_buff_size = size;
}


/*****************************************************************************
 * Global Input Handler
 *
 * All input (as int's) come through this function first, were global-
 * keybindings are handled.
 *
 * After that, if the input was not handled, the input is sent to an input
 * handler for which window has focus.
 ****************************************************************************/
void handle_input(vitunes *v, int input)
{
   switch (input)
   {
      case 12:          /* 12 = CTRL + L */
         wclear(v->ui->player->cwin);
         wclear(v->ui->playlists->cwin);
         wclear(v->ui->medialist->cwin);
         wclear(v->ui->command->cwin);
         clear();
         wrefresh(v->ui->player->cwin);
         wrefresh(v->ui->playlists->cwin);
         wrefresh(v->ui->medialist->cwin);
         wrefresh(v->ui->command->cwin);
         refresh();
         viui_draw_playlists(v);
         viui_draw_medialist(v);
         viui_draw_player(v);
         ui_draw_borders(v->ui);
         wrefresh(v->ui->player->cwin);
         wrefresh(v->ui->playlists->cwin);
         wrefresh(v->ui->medialist->cwin);
         wrefresh(v->ui->command->cwin);
         refresh();
         return;

      case 9:           /* 9 = TAB */
         switch (v->ui->active_window)
         {  case UI_PLAYLISTS:
               ui_set_focus(v->ui, UI_MEDIALIST);
               break;
            case UI_MEDIALIST:
               ui_set_focus(v->ui, UI_PLAYLISTS);
               break;
            case UI_COMMAND:
               ui_set_focus(v->ui, v->ui->previous_window);
               break;
         }
         viui_draw_command(v);
         return;

      case 27:          /* ESCape */
         if (v->ui->active_window == UI_COMMAND)
         {  ui_set_focus(v->ui, v->ui->previous_window);
            viui_draw_command(v);
         }
         return;

      case ':':
         /* Enter command mode */
         if (v->ui->active_window != UI_COMMAND)
            ui_set_focus(v->ui, UI_COMMAND);

         viui_draw_row(v->ui->command, 0, UI_LEFT, "");
         wrefresh(v->ui->command->cwin);
         handle_input_command(v, input);
         return;

      case 'z':
         /* Pause/Un-pause player */
         media_player_pause(v->mplayer);
         return;

      case 's':
         /* Stop player */
         media_player_stop(v->mplayer);
         v->playing_playlist = -1;
         return;

      case 'f':
         /* Seek player forward 10 seconds */
         media_player_seek(v->mplayer, 10);
         return;

      case 'b':
         /* Seek player backard 10 seconds */
         media_player_seek(v->mplayer, -10);
         return;
   }

   /* If reached here, then input was not handled above, so pass it to the
    * appropriate input handler... */
   switch (v->ui->active_window)
   {
      case UI_PLAYLISTS:
         handle_input_playlists(v, input);
         return;

      case UI_MEDIALIST:
         handle_input_medialist(v, input);
         return;

      case UI_COMMAND:
         handle_input_command(v, input);
         return;
   }
}


/*****************************************************************************
 * Playlist-Window Input Handler
 *
 * All input to the playlist window NOT handled by the global handler above
 * is handled here.
 ****************************************************************************/
void handle_input_playlists(vitunes *v, int input)
{
   int   scroll_diff;
   int   xinput;

   switch (input)
   {
      case '\r':  /* ENTER  */
      case '\n':  /* RETURN */
         /* Load current playlist in medialist window */
         v->viewing_playlist = v->ui->playlists->cur_row
                             + v->ui->playlists->voffset;
         v->ui->medialist->voffset  = 0;
         v->ui->medialist->hoffset  = 0;
         v->ui->medialist->cur_row  = 0;
         v->ui->medialist->max_rows = v->library->playlists[v->viewing_playlist]->num_files;
         viui_draw_medialist(v);
         return;

      case 'k':
      case KEY_UP:
         /* Advance current row UP one line */
         ui_window_change_row(v->ui->playlists, -1);
         viui_draw_playlists(v);
         viui_draw_command(v);
         return;

      case 'j':
      case KEY_DOWN:
         /* Advance current row DOWN one line */
         ui_window_change_row(v->ui->playlists, 1);
         viui_draw_playlists(v);
         viui_draw_command(v);
         return;

      case KEY_PPAGE:
      case 21:          /* CRL + U */     /* TODO is this portable? */
         /* Advance screen and current row UP 1/2 the window height */
         scroll_diff = -1 * (v->ui->playlists->height / 2);
         ui_window_vscroll(v->ui->playlists, scroll_diff);
         viui_draw_playlists(v);
         viui_draw_command(v);
         return;

      case KEY_NPAGE:
      case 4:           /* CRL + D */     /* TODO is this portable? */
         /* Advance screen and current row DOWN 1/2 the window height */
         scroll_diff = v->ui->playlists->height / 2;
         ui_window_vscroll(v->ui->playlists, scroll_diff);
         viui_draw_playlists(v);
         viui_draw_command(v);
         return;

      case 'h':
      case KEY_LEFT:
         /* Scroll window horizontally LEFT one character */
         ui_window_hscroll(v->ui->playlists, -1);
         viui_draw_playlists(v);
         viui_draw_command(v);
         return;

      case 'l':
      case KEY_RIGHT:
         /* Scroll window horizontally RIGHT one character */
         ui_window_hscroll(v->ui->playlists, 1);
         viui_draw_playlists(v);
         viui_draw_command(v);
         return;

      case 'G':
         /* Jump to last row in buffer */
         ui_window_vscroll(v->ui->playlists, v->ui->playlists->max_rows);
         v->ui->playlists->cur_row = v->ui->playlists->height - 1;
         if (v->ui->playlists->cur_row >= v->ui->playlists->max_rows)
            v->ui->playlists->cur_row = v->ui->playlists->max_rows - 1;

         viui_draw_playlists(v);
         viui_draw_command(v);
         return;

      case 'y':
         viui_draw_command_mesg(v, "Cannot \"yank\" playlists");
         return;

      case 'd':
         /* Delete the currently selected playlist */

         /* TODO Only 'dd' (single delete) is currently supported */
         while ((xinput = getch()) && xinput == ERR);
         if (xinput != 'd')
            return;

         /* make sure we're not deleting a pseudo playlist */
         int pindex = v->ui->playlists->voffset + v->ui->playlists->cur_row;
         if (pindex == ML_PINDEX_LIBRARY ||  pindex == ML_PINDEX_SRESULTS)
         {  viui_draw_command_error(v, "Can't delete pseudo-playlists");
            return;
         }

         /* build prompt used to ask user if they are sure */
         char *prompt;
         asprintf(&prompt, "Delete playlist \"%s\"? [y/n] ",
            v->library->playlists[pindex]->name);
         if (prompt == NULL)
            err(1, "failed to build prompt");

         /* make sure user wants to delete */
         int answer, ret;
         ret = ui_command_get_yesno(v->ui, prompt, &answer);
         free(prompt);
         if (ret == UI_OK && answer == 1)
         {
            /* remove the playlist */
            ml_remove_playlist(v->library, pindex);
            v->ui->playlists->max_rows--;
            if (v->ui->playlists->cur_row == v->ui->playlists->max_rows)
               v->ui->playlists->cur_row--;

            /* reset display */
            viui_draw_playlists(v);
            if (v->viewing_playlist == pindex)
            {  v->viewing_playlist = ML_PINDEX_LIBRARY;
               v->ui->medialist->cur_row = 0;
               v->ui->medialist->voffset = 0;
               v->ui->medialist->hoffset = 0;
               v->ui->medialist->max_rows = v->library->midb->num_files;
               viui_draw_medialist(v);
            }

            viui_draw_command_mesg(v, "Playlist removed");
         } else
            viui_draw_command_mesg(v, "Delete cancelled");

         return;

   }
}

/*****************************************************************************
 * Medialist-Window Input Handler
 *
 * All input to the medialist window NOT handled by the global handler above
 * is handled here.
 ****************************************************************************/
void handle_input_medialist(vitunes *v, int input)
{
   int   scroll_diff;
   int   xinput;
   int   yankSize;
   char *query;

   switch (input)
   {
      case '\r':
      case '\n':
         /* Start playback of current playlist at current file */
         media_player_set_queue(v->mplayer,
                     v->library->playlists[v->viewing_playlist]);

         v->playing_playlist = v->viewing_playlist;
         v->mplayer->current = v->ui->medialist->cur_row
                             + v->ui->medialist->voffset;
         media_player_play(v->mplayer);

         viui_draw_playlists(v);
         viui_draw_medialist(v);
         return;

      case 'k':
      case KEY_UP:
         /* Advance current row UP one line */
         ui_window_change_row(v->ui->medialist, -1);
         viui_draw_medialist(v);
         viui_draw_command(v);
         return;

      case 'j':
      case KEY_DOWN:
         /* Advance current row DOWN one line */
         ui_window_change_row(v->ui->medialist, 1);
         viui_draw_medialist(v);
         viui_draw_command(v);
         return;

      case KEY_PPAGE:
      case 21:          /* CRL + U */
         /* Advance screen and current row UP 1/2 the window height */
         scroll_diff = -1 * (v->ui->medialist->height / 2);
         ui_window_vscroll(v->ui->medialist, scroll_diff);
         viui_draw_medialist(v);
         viui_draw_command(v);
         return;

      case 4:           /* CRL + D */
      case KEY_NPAGE:
         /* Advance screen and current row DOWN 1/2 the window height */
         scroll_diff = v->ui->medialist->height / 2;
         ui_window_vscroll(v->ui->medialist, scroll_diff);
         viui_draw_medialist(v);
         viui_draw_command(v);
         return;

      case 'h':
      case KEY_LEFT:
         /* Scroll window horizontally LEFT one character */
         ui_window_hscroll(v->ui->medialist, -1);
         viui_draw_medialist(v);
         viui_draw_command(v);
         return;

      case 'l':
      case KEY_RIGHT:
         /* Scroll window horizontally RIGHT one character */
         ui_window_hscroll(v->ui->medialist, 1);
         viui_draw_medialist(v);
         viui_draw_command(v);
         return;

      case 'G':
         /* Jump to last row in buffer */
         ui_window_vscroll(v->ui->medialist, v->ui->medialist->max_rows);
         v->ui->medialist->cur_row = v->ui->medialist->height - 1;
         if (v->ui->medialist->cur_row >= v->ui->medialist->max_rows)
            v->ui->medialist->cur_row = v->ui->medialist->max_rows - 1;

         viui_draw_medialist(v);
         viui_draw_command(v);
         return;

      case '/':
         /* Start a search on the current playlist */

         /* TODO can't search search results */
         if (v->viewing_playlist == ML_PINDEX_SRESULTS)
            break;

         if (ui_command_get_string(v->ui, "/", &query) != UI_OK)
         {  /* user ESC'd out of search */
            viui_draw_row(v->ui->command, 0, UI_LEFT, "");
            wrefresh(v->ui->command->cwin);
         } else
         {
            /* search! */
            ml_search_playlist(v->library, v->viewing_playlist, query);

            /* draw search results */
            v->ui->medialist->voffset = 0;
            v->ui->medialist->hoffset = 0;
            v->ui->medialist->cur_row = 0;
            v->ui->medialist->max_rows = v->library->playlists[ML_PINDEX_SRESULTS]->num_files;
            v->viewing_playlist = ML_PINDEX_SRESULTS;
            viui_draw_medialist(v);
            free(query);

         }
         return;

      case 'y':
         /* skip empty playlist */
         if (v->ui->medialist->cur_row >= v->ui->medialist->max_rows)
            return;

         /* clear any existing copy buffer */
         v->copy_buff_size = 0;

         /* get rest of 'yank' input */
         while ((xinput = getch()) && xinput == ERR);
         if (xinput == 'y')
            copy_buff_set(v, 1);
         else if (xinput == 'G')
         {  copy_buff_set(v, v->ui->medialist->max_rows
               - v->ui->medialist->cur_row - v->ui->medialist->voffset);
         } else if (isdigit(xinput))
         {  if (ui_get_number(v->ui, &yankSize, xinput) == UI_OK)
               copy_buff_set(v, yankSize);
         }

         /* report how many lines were yanked */
         if (v->copy_buff_size > 0)
            viui_draw_command_mesg(v, "%d files yanked", v->copy_buff_size);

         return;

      case 'd':
         /* make sure we're not deleting a pseudo playlist */
         if (v->viewing_playlist == ML_PINDEX_LIBRARY
         ||  v->viewing_playlist == ML_PINDEX_SRESULTS)
         {  viui_draw_command_error(v,
               "Can't alter a pseudo-playlist");
            return;
         }

         /* clear any existing copy buffer */
         v->copy_buff_size = 0;

         /* get the rest of the "delete" input */
         while ((xinput = getch()) && xinput == ERR);
         if (xinput == 'd')
            copy_buff_set(v, 1);
         else if (xinput == 'G')
         {  copy_buff_set(v, v->ui->medialist->max_rows
               - v->ui->medialist->cur_row - v->ui->medialist->voffset);
         } else if (isdigit(xinput))
         {  if (ui_get_number(v->ui, &yankSize, xinput) == UI_OK)
               copy_buff_set(v, yankSize);
         } else
            return;

         /* do the delete */
         if (v->copy_buff_size > 0)
         {  /* get info about where we're deleting from */
            playlist *plist = v->library->playlists[v->viewing_playlist];
            int start = v->ui->medialist->voffset + v->ui->medialist->cur_row;

            int i;
            for (i = 0; i < v->copy_buff_size; i++)
               playlist_remove_file(plist, start);

            v->ui->medialist->max_rows -= v->copy_buff_size;
            if (v->ui->medialist->cur_row >= v->ui->medialist->max_rows)
               v->ui->medialist->cur_row = v->ui->medialist->max_rows - 1;

            viui_draw_medialist(v);
            viui_draw_command_mesg(v, "%d fewer files", v->copy_buff_size);
         }
         return;

      case 'p':
      case 'P':
         /* make sure we're not pasting to a pseudo playlist */
         if (v->viewing_playlist == ML_PINDEX_LIBRARY
         ||  v->viewing_playlist == ML_PINDEX_SRESULTS)
         {  viui_draw_command_error(v,
               "Cannot alter a pseudo-playlist");
            return;
         }

         /* paste any yank'ed files */
         if (v->copy_buff_size == 0)
            viui_draw_command_error(v, "No files to paste!");
         else
         {
            /* determine destination for yank'd files */
            playlist *dest = v->library->playlists[v->viewing_playlist];
            int dest_start;
            if (dest->num_files > 0)
            {  dest_start = v->ui->medialist->voffset
                          + v->ui->medialist->cur_row;

               if (input == 'p') dest_start++;
            } else
               dest_start = 0;

            /* make room in the destination plist */
            int i = dest->num_files + v->copy_buff_size - 1;
            for (; i >= dest_start + v->copy_buff_size; i--)
               dest->files[i] = dest->files[i - v->copy_buff_size];

            dest->num_files += v->copy_buff_size;

            /* copy the files */
            for (i = 0; i < v->copy_buff_size; i++)
               dest->files[dest_start + i] = v->copy_buff[i];

            v->ui->medialist->max_rows += v->copy_buff_size;

            /* update the display */
            viui_draw_medialist(v);
            viui_draw_command_mesg(v, "%d more files", v->copy_buff_size);
         }
         return;
   }
}

/*****************************************************************************
 * Command-Window Input Handler
 *
 * All input to the command window NOT handled by the global handler above
 * is handled here.
 ****************************************************************************/
void handle_input_command(vitunes *v, int input)
{
   char  *cmd;

   switch (input)
   {
      /* enter command mode */
      case ':':   /* SHIFT + colon ( = semicolon) */
         if (ui_command_get_string(v->ui, ":", &cmd) != UI_OK)
         {  /* user cancelled command mode, go back to prev window */
            ui_set_focus(v->ui, v->ui->previous_window);
            viui_draw_row(v->ui->command, 0, UI_LEFT, "");
            wrefresh(v->ui->command->cwin);
            return;
         }

         command_mode_execute(cmd, v);
         ui_set_focus(v->ui, v->ui->previous_window);
         return;
   }
}


