
#include "media_library.h"

/* Function to allocate and return a new media_library struct with default
 * settings.
 *
 * Parameters:
 *    None.
 *
 * Returns:
 *    Pointer to new media_library struct.  It is the responsibility of the
 *    caller to free the returned struct using ml_free().
 */
media_library *ml_new(void)
{
   media_library *ml = malloc(sizeof(media_library));

   ml->midb_file    = NULL;
   ml->playlist_dir = NULL;
   ml->midb         = NULL;

   ml->playlists = calloc(ML_PLAYLISTS_START_SIZE, sizeof(playlist*));
   if (ml->playlists == NULL)
      err(1, "failed to allocate playlists array");

   ml->playlists_capacity = ML_PLAYLISTS_START_SIZE;
   ml->playlists_num      = 0;

   return ml;
}

/* Function to free all of the memory associated with a media_library
 * struct.
 *
 * Parameters:
 *    ml       Pointer to media_library struct you want free()'ed.
 *
 * Returns:
 *    Nothing.
 */
void ml_free(media_library *ml)
{
   /* free the meta information database */
   midb_free(ml->midb);

   /* free all the playlists */
   int i;
   for (i = 0; i < ml->playlists_num; i++)
      playlist_free(ml->playlists[i]);

   free(ml->playlists);
   free(ml->midb_file);
   free(ml->playlist_dir);
   free(ml);
}

/* Function to add a playlist to a media library.
 *
 * Parameters:
 *    ml       Media library to add a playlist to.
 *
 *    p        The playlist being added to the media library.
 *
 * Returns:
 *    Nothing.
 */
void ml_add_playlist(media_library *ml, playlist *p)
{
   /* check to see if we need to resize the array */
   if (ml->playlists_num == ml->playlists_capacity - 1)
   {  int new_size = ml->playlists_capacity + ML_PLAYLISTS_CHUNK_SIZE;
      new_size *= sizeof(playlist*);

      playlist **newp = realloc(ml->playlists, new_size);
      if (newp == NULL)
         err(1, "failed to resize playlists array");

      ml->playlists           = newp;
      ml->playlists_capacity  = new_size;
   }

   ml->playlists[ml->playlists_num++] = p;
}

/* Function to remove a playlist from a media library.
 *
 * Parameters:
 *    ml       Media library to delete a playlist from.
 *
 *    pindex   Index into the media library's playlist array of the playlist
 *             to be removed.
 *
 * Returns:
 *    Nothing.
 */
void ml_remove_playlist(media_library *ml, int pindex)
{
   playlist_delete(ml->playlists[pindex]);

   /* push back */
   int i;
   for (i = pindex + 1; i < ml->playlists_num; i++)
      ml->playlists[i - 1] = ml->playlists[i];

   ml->playlists_num--;
}

/* Function to search a given playlist within the media library using the
 * given query.  For details on how the query is matched against files
 * within a playlist, see the "meta_info_match" function in meta_info.c.
 *
 * Parameters:
 *    ml       Media library to search.
 *
 *    pindex   Index into media library's playlist array of the playlist you
 *             want to search.
 *
 *    query    The search string.
 *
 * Returns:
 *    Nothing.  The search results are stored in the pseudo-playlist at
 *             ml->playlists[ ML_PINDEX_SRESULTS ];
 *
 * Notes:
 *    Remember that calling this function builds/resets the search results
 *    pseudo-playlist.
 */
void ml_search_playlist(media_library *ml, int pindex, char *query)
{
   /* can't search search results */
   if (pindex == ML_PINDEX_SRESULTS)
      return;

   /* tokenize the query */
   int   numTokens = 0;
   char *tokens[1000];  /* XXX */
   char *token;
   
   while ((token = strsep(&query, " \t\n")) != NULL)
   {  if (strlen(token) == 0)
         continue;
      else
         tokens[ numTokens++ ] = token;
   }

   /* search! */
   playlist *plist   = ml->playlists[pindex];
   playlist *results = ml->playlists[ML_PINDEX_SRESULTS];

   int i;
   results->num_files = 0;
   for (i = 0; i < plist->num_files; i++)
   {  if (meta_info_match(plist->files[i], tokens, numTokens))
      {  results->files[ results->num_files ] = plist->files[i];
         results->num_files++;
      }
   }
}

/* Function to load a media library from disk.  The location of the
 * meta-information databae file and the directory containing all playlists
 * are provided.
 *
 * Parameters:
 *    ml       Pointer to an existing media library struct that will be
 *             loaded with the information found in the files specified.
 *
 *    dbfile   Filename of the meta information database.
 *
 *    pdir     Directory containing all of the playlist files.
 *
 * Returns:
 *    Nothing.
 */
void ml_load(media_library *ml, const char *dbfile, const char *pdir)
{
   /* record the file locations provided */
   ml->midb_file    = strdup(dbfile);
   ml->playlist_dir = strdup(pdir);
   if (ml->midb_file == NULL || ml->playlist_dir == NULL)
      err(1, "failed to store media library strings");

   /* load the meta information databaes */
   ml->midb = midb_new();
   midb_load(ml->midb_file, ml->midb);

   /* First load the psuedo-playlists, as they always appear at the beginning
    * of the list.
    */

   /* create the "LIBRARY" pseudo-playlist */
   ml->playlists[ML_PINDEX_LIBRARY] = playlist_new();
   ml->playlists[ML_PINDEX_LIBRARY]->name = strdup("--LIBRARY--");
   ml->playlists[ML_PINDEX_LIBRARY]->num_files = ml->midb->num_files;
   int i;
   for (i = 0; i < ml->midb->num_files; i++)
      ml->playlists[ML_PINDEX_LIBRARY]->files[i] = ml->midb->files[i];

   /* create "SEARCH RESULITS" pseudo-playlist */
   ml->playlists[ML_PINDEX_SRESULTS] = playlist_new();
   ml->playlists[ML_PINDEX_SRESULTS]->name = strdup("<search results>");
   ml->playlists[ML_PINDEX_SRESULTS]->num_files = 0;

   ml->playlists_num = 2;

   /* next, retrieve a list of playlist files in the provided directory */
   char **files;
   int    num_playlists;
   num_playlists = playlist_get_playlist_filenames(pdir, &files);
   num_playlists += ML_NUM_PSEUDO_PLAYLISTS;

   /* load all of the playlists found */
   for (i = ML_NUM_PSEUDO_PLAYLISTS; i < num_playlists; i++)
   {  playlist *p = playlist_load(ml->midb, files[i-2]);
      ml_add_playlist(ml, p);
      free(files[i-2]);
   }
   free(files);
}

