#include "media_player.h"

/* global struct used to monitor the player */
mp_status_t mp_status;

int get_next_song_index(media_player *mp)
{
   if (mp->queue == NULL)
      errx(1, "somwhere, something has gone horribly wrong");

   switch (mp->mode)
   {
      case MP_MODE_LINEAR:
         if (mp->current == mp->queue->num_files - 1)
            return -1;
         else
            return mp->current + 1;
         break;

      case MP_MODE_LOOP:
         if (mp->current == mp->queue->num_files - 1)
            return 0;
         else
            return mp->current + 1;
         break;

      case MP_MODE_RANDOM:
         return rand_r(&(mp->rand_seed)) % mp->queue->num_files;
         break;
   }
   return 0;
}


void media_player_init(media_player *mp)
{
   int   pwrite[2];
   int   pread[2];
   pid_t child_pid;

   /* ignore broken pipe signal -- event is handled elsewhere */
   signal(SIGPIPE, SIG_IGN);

   /* create the pipes */
   if (pipe(pwrite) == -1 || pipe(pread) == -1)
   {  perror("failure creating pipes for media player");
      exit(1);
   }

   /* fork the child */
   child_pid = fork();
   switch (child_pid)
   {  case -1:
         err(1, "couldn't fork() media player");
         break;

      case 0:  /* child process */
         if (close(0) == -1 || close(1) == -1 || close(2) == -1)
            err(1, "failed in fork()'ed child to close stdin/stdout");

         if (close(pwrite[1]) == -1 || close(pread[0]) == -1)
            err(1, "failed in fork()'ed child to close excess pipes");

         if (dup(pwrite[0]) == -1 || dup(pread[1]) == -1)
            err(1, "failed in fork()'ed child dup()'ing pipe");

         /* launch the media player */
         /* TODO: fuck mplayer -- try to find a better msglevel */
         if (execl("/usr/local/bin/mplayer", "mplayer",
               "-slave", "-idle", "-quiet", "-msglevel",
               "cplayer=0:ao=0:vo=0:decaudio=0:decvideo=0:demuxer=0",
               (char*)NULL) == -1)
            err(1, "failed to exec the media player");

         break;
   }

   /* back to the parent... */
   if (close(pwrite[0]) == -1 || close(pread[1]) == -1)
      err(1, "failed to close excess pipes");

   /* set params */
   mp->pid        = child_pid;
   mp->pipe_read  = pread[0];
   mp->pipe_write = pwrite[1];

   mp->queue      = NULL;
   mp->current    = -1;

   mp->mode       = MP_MODE_LOOP;
   mp->rand_seed  = time(0);
   srand(mp->rand_seed);

   /* init status object */
   mp->should_be_playing    = false;
   mp->paused               = false;
   mp->just_started_playing = false;
   mp_status.is_playing = false;
   mp_status.position = 0;

   /* setup read pipe to media player as non-blocking */
   int flags = fcntl(mp->pipe_read, F_GETFL, 0);
   if (flags == -1)
      err(1, "failed to fcntl() the read-pipe to media player (get)");

   if (fcntl(mp->pipe_read, F_SETFL, flags | O_NONBLOCK) == -1)
      err(1, "failed to fcntl() the read-pipe to media player (set)");
}


void media_player_destroy(media_player *mp)
{
   static const char *cmd = "\nquit\n";

   /* reset status object */
   mp_status.is_playing = false;
   mp_status.position = 0;

   /* instruct media player to exit */
   if (write(mp->pipe_write, cmd, strlen(cmd)) == -1)
   {  if (errno == EPIPE)
         return;  /* child is already dead - nothing more to do */
      else
         err(1, "failed to send quit command to media player");
   }

   /* close pipes and wait for media player to die */
   close(mp->pipe_read);
   close(mp->pipe_write);
   waitpid(mp->pid, NULL, 0);
   mp->queue = NULL;
   mp->current = -1;
}

void media_player_set_queue(media_player *mp, playlist *queue)
{
   media_player_stop(mp);

   mp->queue   = queue;
   mp->current = -1;
   if (mp->queue->num_files > 0)
      mp->current = 0;
}

void media_player_play(media_player *mp)
{
   if (mp->queue == NULL)
      errx(1, "media queue is null");

   if (mp->current < 0 || mp->current > mp->queue->num_files)
      errx(1, "media item to play is out of range!");

   /* flush */
   static char response[1000];
   read(mp->pipe_read, &response, 1000);
   read(mp->pipe_read, &response, 1000);

   /* build the command to send to mplayer and send it */
   char *cmd;
   asprintf(&cmd, "\nloadfile \"%s\" 0\nget_time_pos\n",
      mp->queue->files[mp->current]->file);
   write(mp->pipe_write, cmd, strlen(cmd));
   free(cmd);

   mp->just_started_playing = true;
   mp->should_be_playing = true;
   mp->paused            = false;
   mp_status.is_playing = true;
   mp_status.position = 0;
}

void media_player_stop(media_player *mp)
{
   /* XXX FUCK MPLAYER -- yah, this is documented and seems reasonable, but
    * has no affect what-so-ever ...
   static const char *cmd = "\nstop\n";
    * so instead i use the following hack to stop mplayer... load a file
    * that does not exist (well, hopefully there's no file named
    * 'dummy.dummy.dummy').
    */
   static const char *cmd = "\nloadfile dummy.dummy.dummy\n";

   write(mp->pipe_write, cmd, strlen(cmd));

   /* flush */
   static char response[1000];
   read(mp->pipe_read, &response, 1000);
   read(mp->pipe_read, &response, 1000);

   mp->just_started_playing = false;
   mp->should_be_playing = false;
   mp->paused = false;
   mp_status.is_playing = false;
}

void media_player_pause(media_player *mp)
{
   static const char *cmd = "\npause\n";

   write(mp->pipe_write, cmd, strlen(cmd));

   mp->paused = !mp->paused;
}

void media_player_unpause(media_player *mp)
{
   media_player_pause(mp);
}

void media_player_seek(media_player *mp, int seconds)
{
   if (!mp->should_be_playing || mp->paused)
      return;

   char *cmd;
   asprintf(&cmd, "\nseek %i 0\nget_time_pos\n", seconds);
   write(mp->pipe_write, cmd, strlen(cmd));
   free(cmd);
}

void media_player_monitor(media_player *mp)
{
   static const char *info_cmds = "\nget_time_pos\n";
   static const int response_size = 1000;
   static char  response[1000];

   char *response_start;
   int bytes_read;

   /* fuck mplayer, fuck it to hell */
   if (mp->paused) return;

   /* query the media player for info */
   bzero(response, response_size * sizeof(char));
   bytes_read = read(mp->pipe_read, &response, response_size);

   if (bytes_read > 0)
   {  /* sanitize the response and sift through any crap */
      /* TODO there may be multiple ANS_TIME_POSITION responses in the
       * buffer read ... figure how to get to the last one... */
      response[bytes_read + 1] = 0;
      response_start = strstr(response, "ANS_TIME_POSITION");

      /* if an answer to the position query was found, strip it */
      if (response_start != NULL)
      {  if (sscanf(response_start, "ANS_TIME_POSITION=%f",
                &mp_status.position) != 1)
                  errx(1, "mplayer is misbehaving");

         mp->should_be_playing = true;
         mp->paused            = false;
         mp->just_started_playing = false;
         mp_status.is_playing = true;
      }

   } else if (bytes_read == -1)
   {
      if (errno == EAGAIN && mp->should_be_playing && !mp->paused)
      {
         if (mp->just_started_playing)
            mp->just_started_playing = true;
         else if ((mp->current = get_next_song_index(mp)) == -1)
         {  /* reached end of playlist */
            mp->should_be_playing = false;
            mp->paused = false;
            mp_status.is_playing = false;
         } else
            media_player_play(mp);

      }
   }

   /* Re-query the media player for more stats. */
   if (!mp->paused)
      write(mp->pipe_write, info_cmds, strlen(info_cmds));

}

