
#include "meta_info.h"

/* The human-readable names of all of the string-type meta information
 * datums vitunes cares about.
 */
const char *MI_CINFO_NAMES[] = {
   "Artist",
   "Album",
   "Title",
   "Track",
   "Year",
   "Genre"
};


/* Function to create and return a new meta_info struct.  All memory is
 * allocated, but remember that the cinfo array is an array of strings (char*)
 * and those are, of course, not allocated.
 * All pointers are set to NULL defaults, and all numbers to 0.
 *
 * Returns:
 *    Pointer to newly allocated meta_info struct.  If allocation fails,
 *    this function will halt execution.
 *
 * Notes:
 *    It's the responsibility of the caller to free() the struct returned
 *    using the meta_info_free function.
 */
meta_info *meta_info_new(void)
{
   meta_info *mi = (meta_info*) malloc(sizeof(meta_info));
   if (mi == NULL)
      err(1, "meta_info malloc failed");

   mi->file = NULL;
   mi->plength_ms = 0;
   mi->last_updated = 0;

   int i;
   for (i = 0; i < MI_NUM_CINFO; i++)
      mi->cinfo[i] = NULL;

   return mi;
}


/* Function to free() all memory allocated by a given meta_info struct.
 *
 * Parameters:
 *    mi    Pointer to meta_info struct to free.
 *
 * Returns:
 *    nothing.
 */
void meta_info_free(meta_info *mi)
{
   if (mi->file != NULL)
      free(mi->file);

   int i;
   for (i = 0; i < MI_NUM_CINFO; i++)
   {  if (mi->cinfo[i] != NULL)
         free(mi->cinfo[i]);
   }

   free(mi);
}

/* Function to write a meta_info struct to a file stream
 * NOTE: This compliments fread_meta_info below.
 *
 * Parameters:
 *    mi    Pointer to meta_info struct.
 *    fout  File stream
 */
void fwrite_meta_info(meta_info *mi, FILE *fout)
{
   static uint64_t lengths[MI_NUM_CINFO + 1];
   int i;

   /* store all necessary numeric values in the above array */
   lengths[0] = strlen(mi->file);
   for (i = 0; i < MI_NUM_CINFO; i++)
   {  if (mi->cinfo[i] == NULL)
         lengths[i+1] = 0;
      else
         lengths[i+1] = strlen(mi->cinfo[i]);
   }

   /* write */
   fwrite(lengths, sizeof(lengths), 1, fout);
   fwrite(mi->file, sizeof(char), lengths[0], fout);
   for (i = 0; i < MI_NUM_CINFO; i++)
   {  if (mi->cinfo[i] != NULL)
         fwrite(mi->cinfo[i], sizeof(char), lengths[i+1], fout);
   }
   fwrite(&(mi->plength_ms),   sizeof(uint64_t), 1, fout);
   fwrite(&(mi->last_updated), sizeof(time_t),   1, fout);

   fflush(fout);
}

/* Function to read a meta_info struct from a file stream
 * NOTE: This compliments fwrite_meta_info above.
 *
 * Parameters:
 *    mi    Pointer to EXISTING meta_info struct where info will be stored
 *    fin   File stream
 */
void fread_meta_info(meta_info *mi, FILE *fin)
{
   static uint64_t lengths[MI_NUM_CINFO + 1];
   int i;

   /* first read all necessary numeric values */
   fread(lengths, sizeof(lengths), 1, fin);

   /* allocate all needed space in the meta_info struct, and zero */
   if ((mi->file = calloc(lengths[0] + 1, sizeof(char))) == NULL)
      err(1, "failure to malloc meta_info file");

   bzero(mi->file, sizeof(char) * (lengths[0] + 1));
   for (i = 0; i < MI_NUM_CINFO; i++)
   {  if (lengths[i+1] > 0)
      {  if ((mi->cinfo[i] = calloc(lengths[i+1] + 1, sizeof(char))) == NULL)
            err(1, "failure to malloc meta_info CINFO");

         bzero(mi->cinfo[i], sizeof(char) * (lengths[i+1] + 1));
      }
   }

   /* read */
   fread(mi->file, sizeof(char), lengths[0], fin);
   for (i = 0; i < MI_NUM_CINFO; i++)
   {  if (lengths[i+1] > 0)
         fread(mi->cinfo[i], sizeof(char), lengths[i+1], fin);
   }
   fread(&(mi->plength_ms),   sizeof(uint64_t), 1, fin);
   fread(&(mi->last_updated), sizeof(time_t),   1, fin);
}

/* Simple function to help with pulling ID3 information.  libid3tag always
 * "lifts" the information returned to a ID3v2.4 format, which uses
 * unicode for textual meta info (which is all we're interested in).
 * Given an id3 tag, and the name of a "frame" (meta data atom) to pull,
 * this extracts the meta information, if there, and converts to utf8 (char).
 *
 * Parameters:
 *    tag         Pointer to id3_tag struct
 *    frame_id    String representing name of tag to extract
 *
 * Returns:
 *    IF the given frame exists in the tag, then a newly allocated string
 *    in utf8 (char) is returned.  Otherwise, or in case of any errors,
 *    NULL is returned.
 *
 * Notes:
 *    It's the responsibility of the caller to free() the string returned.
 */
char *id3_pull_tag2char(struct id3_tag *tag, char *frame_id)
{
   struct id3_frame *frame;
   id3_ucs4_t const *strUnicode; /* unicode version of the data */
   id3_utf8_t *str;        /* good 'ol char version */

   /* check to see if frame with given ID exists in tag */
   if ((frame = id3_tag_findframe(tag, frame_id, 0)) == NULL)
      return NULL;

   /* Extract the second field (with the data) and see how many strings are
    * there.
    * NOTE: The only frames we're interested in always have two fields.
    * The first is the encoding, and the second is the actual text.
    */
   if (id3_field_getnstrings(&frame->fields[1]) < 1)
   {  id3_frame_delete(frame);
      return NULL;
   }

   /* Extract just the first field */
   strUnicode = id3_field_getstrings(&frame->fields[1], 0);
   if (strUnicode == NULL)
   {  id3_frame_delete(frame);
      return NULL;
   }

   /* convert to char and return ... also free stuff */
   str = id3_ucs4_utf8duplicate(strUnicode);
   if (str == NULL)
   {  id3_frame_delete(frame);
      return NULL;
   }

   id3_frame_delete(frame);
   return str;
}

/*****************************************************************************
 * Each of the following functions is of the form:
 *
 *    int meta_extract_FORMAT(char *filename, meta_info *info)
 *
 * Where FORMAT is the type of meta information the function attempts to
 * extract (i.e. "id3", "mp4", "vorbis", etc.).
 *
 * For each function, the parameters and return values are interpretted the
 * same.
 *
 * Parameters:
 *    filename       String containing the file to try and extract the info
 *                   from.
 *    mi             Pointer to ALREADY ALLOCATED meta_info struct that will
 *                   be filled with any meta information found in the file.
 * Returns:
 *    IF meta information is found and extracted succesfully, 0 is returned.
 *    IF no meta information was found, but otherwise everything was
 *    successfull, then MI_ERR_NOINFO is returned.
 *    Otherwise, some error occurred while attempting to extract the info,
 *    and an appropriate error is returned.  Possible errors are:
 *
 *          MI_ERR_FOPEN      Error opening file.
 *
 *          MI_ERR_MALLOC     Error trying allocate memory
 *
 *          MI_ERR_UNKNOWN    Returned when the unerlying API provides no
 *                            documentation as to why something went wrong.
 *                            Sadly, this is the most common.
 *
 ****************************************************************************/

/* Extractor for ID3 meta information */
int meta_extract_id3(char *filename, meta_info *mi)
{
   struct id3_file  *file;
   struct id3_tag   *tag;

   /* open file and get id3 tag */
   file = id3_file_open(filename, ID3_FILE_MODE_READONLY);
   if (file == NULL)
      return MI_ERR_FOPEN;

   tag = id3_file_tag(file);
   if (tag == NULL)
   {  id3_file_close(file);
      return MI_ERR_UNKNOWN;
   }

   /* pull all info */
   mi->cinfo[MI_CINFO_ARTIST] = id3_pull_tag2char(tag, ID3_FRAME_ARTIST);
   mi->cinfo[MI_CINFO_ALBUM]  = id3_pull_tag2char(tag, ID3_FRAME_ALBUM);
   mi->cinfo[MI_CINFO_TITLE]  = id3_pull_tag2char(tag, ID3_FRAME_TITLE);
   mi->cinfo[MI_CINFO_TRACK]  = id3_pull_tag2char(tag, ID3_FRAME_TRACK);
   mi->cinfo[MI_CINFO_YEAR]   = id3_pull_tag2char(tag, ID3_FRAME_YEAR);
   mi->cinfo[MI_CINFO_GENRE]  = id3_pull_tag2char(tag, ID3_FRAME_GENRE);

   /* The genre pulled above is a number (as a string) that's the index
    * into the ID3 list of genres.  here, decode that number into full
    * string.
    */
   if (mi->cinfo[MI_CINFO_GENRE] != NULL)
   {  long long gindex;
      const char *errstr;

      /* note that the range is here what libid3tag can handle */
      gindex = strtonum(mi->cinfo[MI_CINFO_GENRE], 0, 147, &errstr);
      if (errstr != NULL)  /* check for invalid genre number */
      {  free(mi->cinfo[MI_CINFO_GENRE]);
         mi->cinfo[MI_CINFO_GENRE] = NULL;
      } else
      {  id3_ucs4_t const *gstr = id3_genre_index((unsigned int)gindex);
         free(mi->cinfo[MI_CINFO_GENRE]);
         mi->cinfo[MI_CINFO_GENRE] = id3_ucs4_utf8duplicate(gstr);
      }
   }

   /* TODO figure out how to get the total time of a track */
   mi->plength_ms = 0;

   /* cleanup */
   id3_tag_delete(tag);
   id3_file_close(file);

   /* check if we found anything */
   int i;
   for (i = 0; i < MI_NUM_CINFO; i++)
   {  if (mi->cinfo[i] != NULL)
         return 0;
   }

   return MI_ERR_NOINFO;
}

/* Extractor for MP4/AAC meta information */
int meta_extract_mp4(char *filename, meta_info *mi)
{
   uint16_t track1, track2;
   MP4FileHandle file;

   /* open file */
   file = MP4Read(filename, 0);
   if (file == MP4_INVALID_FILE_HANDLE)
      return MI_ERR_FOPEN;

   /* pull data.
    * NOTE: These functions will allocate each str if such a property is
    * found.  otherwise, they set the string to NULL.
    */
   MP4GetMetadataName(file,   &(mi->cinfo[MI_CINFO_TITLE]));
   MP4GetMetadataArtist(file, &(mi->cinfo[MI_CINFO_ARTIST]));
   MP4GetMetadataAlbum(file,  &(mi->cinfo[MI_CINFO_ALBUM]));
   MP4GetMetadataYear(file,   &(mi->cinfo[MI_CINFO_YEAR]));
   MP4GetMetadataTrack(file,  &track1, &track2);
   if (track1 > 0)
   {  if (asprintf(&(mi->cinfo[MI_CINFO_TRACK]), "%d", track1) == -1)
         errx(1, "failure to copy mp4 track info for '%s'", filename);
   }

   /* XXX For some reason, this sometimes bombs when the genre in the m4a
    * file is specified using an ID3 style genre-id (number).
    * no clue why... been staring at the code for ages.
    * as such, no genre from mp4 files for now
    * TODO re-dig-into libmp4v2 to see why this is happening.
   MP4GetMetadataGenre(file,  &(info->genre));
    */

   /* get the duration in ms
    * NOTE: an MP4 can contain multiple "tracks", each of which as their
    * own length/time.  We take the maximum of all track lengths.
    */ 
   mi->plength_ms = 0;
   int i;
   uint32_t numTracks = MP4GetNumberOfTracks(file, NULL, 0);
   for (i = 0; i < numTracks; i++)
   {  MP4TrackId track      = MP4FindTrackId(file, i, NULL, 0);
      MP4Duration numTicks  = MP4GetTrackDuration(file, track);
      uint64_t time = MP4ConvertFromTrackDuration(file, track, numTicks,
         MP4_USECS_TIME_SCALE);

      if (time > mi->plength_ms)
         mi->plength_ms = time;
   }

   /* cleanup */
   MP4Close(file);

   /* check if we found anything */
   for (i = 0; i < MI_NUM_CINFO; i++)
   {  if (mi->cinfo[i] != NULL)
         return 0;
   }

   return MI_ERR_NOINFO;
}

/* Extractor for OGG/VORBIS meta information */
int meta_extract_vorbis(char *filename, meta_info *mi)
{
   OggVorbis_File file;
   vorbis_comment *vcomments;

   /* open file and pull comments */
   if (ov_fopen(filename, &file) < 0)
      return MI_ERR_FOPEN;

   vcomments = ov_comment(&file, -1);
   if (vcomments == NULL)
   {  ov_clear(&file);
      return MI_ERR_UNKNOWN;
   }

   /* pull info */
   int i;
   for (i = 0; i < vcomments->comments; i++)
   {
      char *value = strchr(vcomments->user_comments[i], '=') + 1;

      if (strncasecmp(vcomments->user_comments[i], "artist", 6) == 0)
         asprintf(&(mi->cinfo[MI_CINFO_ARTIST]), "%s", value);
      else if (strncasecmp(vcomments->user_comments[i], "album", 5) == 0)
         asprintf(&(mi->cinfo[MI_CINFO_ALBUM]), "%s", value);
      else if (strncasecmp(vcomments->user_comments[i], "title", 5) == 0)
         asprintf(&(mi->cinfo[MI_CINFO_TITLE]), "%s", value);
      else if (strncasecmp(vcomments->user_comments[i], "tracknumber", 11) == 0)
         asprintf(&(mi->cinfo[MI_CINFO_TRACK]), "%s", value);
      else if (strncasecmp(vcomments->user_comments[i], "date", 4) == 0)
         asprintf(&(mi->cinfo[MI_CINFO_YEAR]), "%s", value);
      else if (strncasecmp(vcomments->user_comments[i], "genre", 5) == 0)
         asprintf(&(mi->cinfo[MI_CINFO_GENRE]), "%s", value);
   }

   /* get the time */
   mi->plength_ms = 0;
   double time = ov_time_total(&file, -1);
   if (time != OV_EINVAL)
      mi->plength_ms = (uint64_t)(time * 1000000);

   /* cleanup */
   ov_clear(&file);
   /* XXX does the vorbis_comment struct need to be free()'d?  Or is it
    * setup to point to existing structs in the oggfile struct?
    * Documentation is no help, and I see not ov_comment_free() function
    * in the library.
    */

   /* check if we found anything */
   for (i = 0; i < MI_NUM_CINFO; i++)
   {  if (mi->cinfo[i] != NULL)
         return 0;
   }

   return MI_ERR_NOINFO;
}

/* The main top-level function of this module.  It's also the only function
 * (aside from meta_info_new and meta_info_free) used outside of this module.
 * Given a the name of a file, this function will attempt to extract any
 * meta information from the file and return a pointer to a newly allocated
 * meta_info struct.  If unsuccessfull, NULL is returned.
 *
 * Parameters:
 *    filename    Name of file to extract info from.
 *
 * Returns:
 *    If successfull in extracting meta information, a pointer to a newly
 *    allocated meta_info struct is returned.
 *    If unsuccessfull, for ANY reason, NULL is returned.
 *
 * Notes:
 *    It's the responsibility of the caller to free the meta_info struct
 *    returned using the meta_info_free function.
 *
 *    The filename given will be expanded using realpath(3).  If that fails,
 *    a warning is displayed to stderr and NULL is returned, but processing
 *    will continue.
 *
 *    If a malloc() within this function fails, the program is halted
 *    immediately.
 *
 *    Any other error (namely within the meta_extract_FORMAT functions) are
 *    simply ignored, and there is currently no way to querry what/where the
 *    error was. TODO
 */
meta_info *meta_extract(const char *filename)
{
   static char fullname[PATH_MAX];

   /* create meta info struct */
   meta_info *mi = meta_info_new();

   /* store full filename in meta_info struct */
   bzero(fullname, sizeof(fullname));
   if (realpath(filename, fullname) == NULL)
      err(1, "couldn't resolve '%s' to full path name", filename);

   if (asprintf(&(mi->file), "%s", fullname) == -1)
      errx(1, "failure to copy filename '%s' for meta info", fullname);

   time(&mi->last_updated);

   /* check each of the formats supported in an arbitrary order.
    * if/when any of them find meta info, return immediately.
    * NOTE: if any of the extraction function encounters any error, we
    * simply keep trying every other function.
    */

   if (meta_extract_id3(fullname, mi) == 0)
      return mi;

   if (meta_extract_vorbis(fullname, mi) == 0)
      return mi;

   if (meta_extract_mp4(fullname, mi) == 0)
      return mi;

   /* if we've reached here, no meta info found.  free the meta_info
    * struct and return null.
    */
   meta_info_free(mi);
   return NULL;
}

/* Function to see if a given meta info object matches a pre-tokenized
 * query.
 */
bool meta_info_match(const meta_info *mi, char *tokens[], int numTokens)
{
   int i, j;
   for (i = 0; i < numTokens; i++)
   {
      bool tokenMatches = false;

      if (strcasestr(mi->file, tokens[i]) != NULL)
         tokenMatches = true;

      for (j = 0; !tokenMatches && j < MI_NUM_CINFO; j++)
      {  if (mi->cinfo[j] == NULL) continue;

         if (strcasestr(mi->cinfo[j], tokens[i]) != NULL)
            tokenMatches = true;
      }

      if (!tokenMatches)
         return false;
   }

   return true;
}
