#ifndef META_INFO_H
#define META_INFO_H

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <err.h>

/* NOTE: This is required since mp4.h (below) tries to define 'bool' as a
 * typedef, which ncurses.h (elsewhere) also does.  They conflict, and
 * cause headaches.
 */
#include <stdbool.h> 

#include <id3tag.h>
#include <vorbis/vorbisfile.h>
#include <mp4.h>

/* error codes used in this module */
#define MI_ERR_FOPEN    -1
#define MI_ERR_MALLOC   -2
#define MI_ERR_UNKNOWN  -3
#define MI_ERR_NOINFO   -4

/* fixed size of the cinfo member of the meta_info struct, containain all
 * textual meta information of a media file.
 */
#define MI_NUM_CINFO    6

/* these are the indices into the cinfo member of the meta_info struct */
#define MI_CINFO_ARTIST 0
#define MI_CINFO_ALBUM  1
#define MI_CINFO_TITLE  2
#define MI_CINFO_TRACK  3
#define MI_CINFO_YEAR   4
#define MI_CINFO_GENRE  5

/* Array of human-readable names of each CINFO meta datum represented in the
 * "cinfo" part of the meta_info struct.
 */
extern const char *MI_CINFO_NAMES[MI_NUM_CINFO];

typedef struct {
   char       *file;                   /* filename of file itself */
   char       *cinfo[MI_NUM_CINFO];    /* character meta info array */
   uint64_t    plength_ms;             /* play length in MICROseconds */
   time_t      last_updated;           /* time info was extracted */
} meta_info;

/* The key struct of this module, storing all of the meta-information about
 * media files we are interested in.
 * This includes the following:
 *    1) file        (string) The location of the file stored on disk.
 *                   NOTE: this is the full, expanded pathname (realpath(3)).
 *    2) cinfo       (char*[]) array of strings reprsenting each of the textual
 *                   meta datums we are interested in.  The array is of fixed
 *                   size, and each index corresponds to a specific datum.
 *                   The indices are defined above as MI_CINFO_*
 *    3) plength_ms  (num) The play-time of the media file in microseconds.
 *    4) last_updated (time) time since info was extracted from file.
 */

/* functions to create and free meta_info structs */
meta_info *meta_info_new(void);
void meta_info_free(meta_info *info);

/* functions to read/write a meta_info object to a file */
void fwrite_meta_info(meta_info *mi, FILE *fout);
void fread_meta_info(meta_info *mi,  FILE *fin);

/* function to see if a given meta-info object matches a given query
 * (already token-ized) */
bool meta_info_match(const meta_info *mi, char *tokens[], int numTokens);

/* main function used to extract meta info from a media file */
meta_info* meta_extract(const char *filename);

/*****************************************************************************
 * The following are functions should NOT be used outside of this module.
 ****************************************************************************/

/* Functions to extract meta_info for individual media formats.  For each of
 * these, the meta_info* passed must already be allocated.
 */
int meta_extract_id3(char *filename, meta_info *info);
int meta_extract_mp4(char *filename, meta_info *info);
int meta_extract_vorbis(char *filename, meta_info *info);

#endif
