
#include "meta_info_db.h"

extern int errno;

/* Function to allocate and return a new meta_info_db struct, initialized to
 * an empty state.
 *
 * Parameters:
 *    None
 *
 * Returns:
 *    Pointer to a new meta_info_db struct.  It is the responsibility of the
 *    caller to free that struct using midb_free().
 */
meta_info_db *midb_new(void)
{
   meta_info_db *midb = malloc(sizeof(meta_info_db));
   if (midb == NULL)
      err(1, "failed to allocate meta info DB.");

   midb->num_files = 0;
   return midb;
}

/* Function to free all of the memory associated with a meta informationd DB
 * object.
 *
 * Parameters:
 *    midb     Pointer to the meta_info_db struct to free.
 *
 * Returns:
 *    Nothing
 */
void midb_free(meta_info_db *midb)
{
   int i;
   for (i = 0; i < midb->num_files; i++)
      free(midb->files[i]);

   free(midb);
}

/* Function to read in a meta_info_db from the file specified, and store it
 * in the passed meta_info_db object.
 *
 * Parameters:
 *    dbfile         Filename of the database file.
 *
 *    midb           Pointer to EXISTING meta_info_db struct where the
 *                   database will be loaded into.
 *
 * Returns:
 *    Nothing
 */
void midb_load(const char *dbfile, meta_info_db *midb)
{
   meta_info *mi;

   /* open db file */
   FILE *fin = fopen(dbfile, "r");
   if (fin == NULL)
      err(1, "failed to open DB at '%s'", dbfile);

   /* read in records adding them to the db */
   midb->num_files = 0;
   while (!feof(fin))
   {  mi = meta_info_new();
      fread_meta_info(mi, fin);
      if (feof(fin))
         meta_info_free(mi);
      else
         midb_add_record(midb, mi);
   }

   fclose(fin);
}

/* Function to store a meta_info_db object to a file.
 *
 * Parameters:
 *    dbfile      Filename where to save the meta_info_db
 *
 *    midb        Pointer to the meta_info_db object to save to file.
 *
 * Returns:
 *    Nothing.
 */
void midb_save(const char *dbfile, const meta_info_db *midb)
{
   /* open db file */
   FILE *fout = fopen(dbfile, "w");
   if (fout == NULL)
      err(1, "couldn't open/create DB '%s'", dbfile);

   /* write each record to the file */
   int i;
   for (i = 0; i < midb->num_files; i++)
      fwrite_meta_info(midb->files[i], fout);

   fclose(fout);
}

/* Function to dump the contents of a meta_info_db struct to a file in a
 * semi-readable (but easily parsable) format.
 *
 * Parameters:
 *    midb        Pointer to meta_info_db to dump.
 *
 *    fout        File stream to dump the meta_info_db contents to.
 *
 * Returns:
 *    Nothing.
 */
void midb_flush(const meta_info_db *midb, FILE *out)
{
   int i;
   for (i = 0; i < midb->num_files; i++)
   {
      /* output the filename first */
      fprintf(out, "%s\n", midb->files[i]->file);

      /* output all of the character-info fields */
      int j;
      for (j = 0; j < MI_NUM_CINFO; j++)
      {  fprintf(out, "\t%6s: %s\n", MI_CINFO_NAMES[j],
            midb->files[i]->cinfo[j]);
      }

      /* output the time */
      fprintf(out, "\t  Time: %f\n",
         (double)midb->files[i]->plength_ms / 1000000.0);

      fflush(out);
   }

}

/* Function to add a new meta-info record to the database.
 *
 * Parameters:
 *    midb        Pointer to the database we're adding a record to.
 *
 *    file        meta_info object (record) to add.
 *
 * Returns:
 *    Nothing.
 */
void midb_add_record(meta_info_db *midb, meta_info *file)
{
   if (midb->num_files == MAX_NUM_SONGS_PER_PLAYLIST - 1)
      errx(1, "Too many songs in meta information database.");

   midb->files[ midb->num_files++ ] = file;
}

/* Function ot update a record within a database.
 *
 * Parameters:
 *    midb        The database being updated.
 *
 *    record      Index into array of the record to update.
 *
 *    newRecord   The new value to store at the index specified.
 *
 * Returns:
 *    Nothing.
 */
void midb_update_record(meta_info_db *midb, int record, meta_info *newRecord)
{
   meta_info_free(midb->files[record]);
   midb->files[record] = newRecord;
}

/* Function to remove a record from a meta_info database.
 *
 * Parameters:
 *    midb        Pointer to database we're removing a record from.
 *
 *    index       Index into the array of the record you want removed.
 *
 * Returns:
 *    Nothing.
 */
void midb_remove_record(meta_info_db *midb, int index)
{
   int i;
   for (i = index; i < midb->num_files; i++)
      midb->files[i] = midb->files[i+1];

   midb->num_files--;
}

/* Function to query a meta-info database for a record with the provided
 * filename.
 *
 * Parameters:
 *    midb        Database to query.
 *    filename    Filename of record to search for.
 *
 * Returns:
 *    If such a record is found, the index is returned.  Otherwise, -1 is
 *    returned.
 */
int midb_query_by_filename(const meta_info_db *midb, const char *filename)
{
   int i;
   for (i = 0; i < midb->num_files; i++)
   {  if (strcmp(filename, midb->files[i]->file) == 0)
         return i;
   }

   return -1;
}

/* Function to create a new database -or- update an existing database,
 * by scanning a NULL-terminated array of files/directories for media files
 * with meta-information.  Directories are searched recursively.
 *
 * The newly created -or- updated database is saved to the file specified.
 *
 * Parameters:
 *    dbfile      Filename where to save the database that's created.
 *                If a database already exists in this file, it is
 *                first loaded and the search continues using this
 *                existing database.
 *
 *    dirlist     A NULL-terminated array of files/directories that will
 *                be search for media files.
 *
 * Returns:
 *    Nothing.
 */
void midb_create(const char *dbfile, char *dirlist[])
{
   /* various counters used to keep track of progress */
   unsigned int count_removed_lost_info = 0;
   unsigned int count_updated = 0;
   unsigned int count_skipped_no_info = 0;
   unsigned int count_skipped_dir = 0;
   unsigned int count_skipped_error = 0;
   unsigned int count_added = 0;


   /* STEP 1.  Create a new empty database */
   meta_info_db *midb = midb_new();


   /* STEP 2.  Check if a database already exists at the given filename.
    *          If so, load it, and we'll add files to it.
    *          Otherwise, we work with an empty enitial database.
    */
   struct stat dbinfo;
   switch (stat(dbfile, &dbinfo))
   {  case 0:  /* DB DOES exist -- load it */
         midb_load(dbfile, midb);
         break;
      case -1: /* DB does NOT exist -or- there was an error during stat() */
         switch (errno)
         {  case ENOENT:
               /* file simply doesn't exist -- that's fine */
               break;
            default:
               err(1, "database file '%s'", dbfile);
         }
         break;
   }


   /* STEP 3.  Start searching the files/directories specified for any
    *          media files with meta information...
    */
   FTS *fts = fts_open(dirlist, FTS_LOGICAL | FTS_NOCHDIR, NULL);
   if (fts == NULL)
      err(1, "couldn't search dir heirarchy");

   meta_info *mi;
   char       fullname[PATH_MAX];
   FTSENT    *ftsent;

   while ((ftsent = fts_read(fts)) != NULL)
   {  /* What type of file is this... */
      switch (ftsent->fts_info)
      {
         case FTS_D:    /* TYPE: directory (going in) */
            printf("DIRECTORY: %s\n", ftsent->fts_path);
            break;

         case FTS_DP:   /* TYPE: directory (coming out) */
            break;

         case FTS_DNR:  /* TYPE: unreadable directory */
            printf("(SKIPPED, CAN'T READ DIR): %s\n", ftsent->fts_accpath);
            count_skipped_dir++;
            break;

         case FTS_NS:   /* TYPE: file/dir that couldn't be stat(2) */
         case FTS_ERR:  /* TYPE: other error */
            printf("(SKIPPED, ERROR) %s\n", ftsent->fts_path);
            count_skipped_error++;
            break;

         case FTS_F:    /* TYPE: regular file */

            /* get the full name for the file */
            if (realpath(ftsent->fts_accpath, fullname) == NULL)
               err(1, "failed to get full path for \"%s\"",
                  ftsent->fts_accpath);

            /* check if the file already exists in the db */
            int index = midb_query_by_filename(midb, fullname);
            if (index != -1)  /* file DOES exist in DB... */
            {
               /* check if the file has been modified since it was added */
               if (ftsent->fts_statp->st_mtim.tv_sec >
                   midb->files[index]->last_updated)
               {  /* it has been modified... update the record */
                  mi = meta_extract(ftsent->fts_accpath);
                  if (mi == NULL)   /* had info before, now it doesn't */
                  {  midb_remove_record(midb, index);
                     printf("(REMOVED, NO INFO) %s\n", ftsent->fts_accpath);
                     count_removed_lost_info++;
                  } else            /* still has info, update record */
                  {  midb_update_record(midb, index, mi);
                     printf("(UPDATED) %s\n", ftsent->fts_accpath);
                     count_updated++;
                  }
               }

            } else            /* file does NOT exist in DB... */
            {
               mi = meta_extract(ftsent->fts_accpath);
               if (mi == NULL)   /* file has no info */
               {  printf("(SKIPPED, NO INFO) %s\n", ftsent->fts_accpath);
                  count_skipped_no_info++;
               }
               else              /* file does have info */
               {  midb_add_record(midb, mi);
                  printf("(ADDED) %s\n", ftsent->fts_accpath);
                  count_added++;
               }
            }
      }
   }

   /* save to file */
   midb_save(dbfile, midb);

   /* output some of our stats */
   printf("--------------------------------------------------\n");
   printf("Results of scanning directories...\n");
   printf("%9d files added\n", count_added);
   printf("%9d files updated\n", count_updated);
   printf("%9d files removed (no longer have meta-info)\n",
      count_removed_lost_info);
   printf("%9d files skipped (no info)\n", count_skipped_no_info);
   printf("%9d files skipped (other error)\n", count_skipped_error);
   printf("%9d directories skipped (couldn't read)\n", count_skipped_dir);
}

/* Function to load a database and scan all the records to see if the files
 * they reference have been removed, modified, etc.  Any changed files are
 * updated within the database.
 *
 * Parameters:
 *    dbfile         Filename of database to update.
 *
 * Returns:
 *    Nothing.
 */
void midb_update(const char *dbfile)
{
   /* various counters used to keep track of our progress */
   unsigned int count_files_removed_file_gone = 0;
   unsigned int count_files_removed_meta_gone = 0;
   unsigned int count_files_updated = 0;
   unsigned int count_files_error = 0;

   /* load the database */
   meta_info_db *midb = midb_new();
   midb_load(dbfile, midb);

   int i;
   for (i = 0; i < midb->num_files; i++)
   {
      char *filename = midb->files[i]->file;

      struct stat sb;
      if (stat(filename, &sb) == -1)
      {  /* file was removed -or- stat() failed */
         if (errno == ENOENT)
         {  midb_remove_record(midb, i);
            printf("(REMOVED, FILE GONE) %s\n", filename);
            count_files_removed_file_gone++;
         } else
         {  printf("(KEPT, BUT COULDN'T STAT), %s\n", filename);
            count_files_error++;
         }

      } else
      {  /* file still exists... check if it has been modified */
         if (sb.st_mtim.tv_sec > midb->files[i]->last_updated)
         {  meta_info *mi = meta_extract(filename);
            if (mi == NULL)
            {  midb_remove_record(midb, i);
               printf("(REMOVED, META-INFO GONE) %s\n", filename);
               count_files_removed_meta_gone++;
            } else
            {  midb_update_record(midb, i, mi);
               printf("(UPDATED) %s\n", filename);
               count_files_updated++;
            }
         }
      }
   }

   /* save to file */
   midb_save(dbfile, midb);

   /* output some of our stats */
   printf("--------------------------------------------------\n");
   printf("Results of updating database...\n");
   printf("%9d files updated\n", count_files_updated);
   printf("%9d files removed (file no longer exists)\n",
      count_files_removed_file_gone);
   printf("%9d files removed (meta-info gone)\n",
      count_files_removed_meta_gone);
   printf("%9d files with errors (couldn't stat, but kept)\n",
      count_files_error);
}

