
#include "playlist.h"

/* Allocate a new playlist and return a pointer to it.  It is the
 * responsibility of the caller to free the returned playlist using
 * playlist_free().
 *
 * Parameters:
 *    None.
 *
 * Returns:
 *    Pointer to a newly allocated playlist structure.
 */
playlist *playlist_new(void)
{
   playlist *p = malloc(sizeof(playlist));
   if (p == NULL)
      err(1, "failed to allocate playlist");

   p->filename  = NULL;
   p->name      = NULL;
   p->num_files = 0;

   return p;
}

/* Free all of the memory consumed by a given playlist.
 *
 * Parameters:
 *    p     Pointer to the playlist you wish to be free'ed.
 */
void playlist_free(playlist *p)
{
   if (p->filename != NULL) free(p->filename);
   if (p->name != NULL) free(p->name);
   free(p);
}

/* Duplicate an existing playlist, returning a pointer to the newly allocated
 * playlist.  The filename and name of the new playlist must be specified.
 *
 * Parameters:
 *    original    Pointer to existing playlist you wish to create a copy of.
 *
 *    filename    String to set the filename of the NEW playlist to.
 *    name        String to set the name of the NEW playlist to.
 *
 * Returns:
 *    Pointer to a newly allocated copy of the original playlist.  It is
 *    the responsibility of the caller to free that playlist with
 *    playlist_free().
 */
playlist *playlist_dup(const playlist *original, const char *filename,
   const char *name)
{
   /* create new playlist and copy simple members */
   playlist *newplist = playlist_new();
   newplist->num_files = original->num_files;
   newplist->name      = strdup(name);
   newplist->filename  = strdup(filename);

   if (newplist->name == NULL || newplist->filename == NULL)
      err(1, "unable to duplicate playlist names");

   /* copy all of the files */
   int i;
   for (i = 0; i < original->num_files; i++)
      newplist->files[i] = original->files[i];

   return newplist;
}

/* Loads a playlist from the provided file.  The files within the playlist
 * are compared against the given meta-information-database to see if they
 * exist there.  If they do, the corresponding entry in the playlist
 * structure built is simply a pointer to the existing entry.
 * Otherwise, that file's meta information is set to NULL and the file is
 * copied (allocated) in the playlist structure.
 * A newly allocated playlist is returned.
 *
 * Parameters:
 *    midb        Pointer the meta information database.
 *
 *    filename    The name of the file that the playlist will be loaded from.
 *
 * Returns:
 *    A pointer to a newly allocated playlist, loaded from the given file.
 *    It is the responsibility of the caller to free the playlist using
 *    playlist_free().
 */
playlist *playlist_load(meta_info_db *midb, const char *filename)
{

   /* open file */
   FILE     *fin;
   if ((fin = fopen(filename, "r")) == NULL)
      err(1, "failed to open playlist '%s'", filename);

   /* create playlist and setup */
   playlist *p = playlist_new();
   p->filename = strdup(filename);
   p->name     = strdup(basename(filename));
   if (p->filename == NULL || p->name == NULL)
      err(1, "failed to allocate info for playlist '%s'", filename);

   /* hack to remove '.playlist' from name */
   char *period = strrchr(p->name, '.');
   *period = '\0';

   /* read each line from the file and copy into playlist object */
   static char media_file[PATH_MAX + 1];
   while (fgets(media_file, PATH_MAX, fin) != NULL)
   {
      if (p->num_files == MAX_NUM_SONGS_PER_PLAYLIST - 1)
      {  warnx("Too many songs in playlist file '%s'\n", filename);
         break;
      }

      /* sanitize */
      media_file[strcspn(media_file, "\n")] = '\0';

      /* check if file exists in the meta info. db */
      int index = midb_query_by_filename(midb, media_file);
      if (index >= 0)   /* file DOES exist */
         p->files[p->num_files] = midb->files[index];
      else              /* file does NOT exist */
      {
         /* create empty meta-info object with just the file name */
         meta_info *mi = meta_info_new();
         mi->file = strdup(media_file);
         if (mi->file == NULL)
            err(1, "unable to copy playlist file '%s'", media_file);

         /* add new record to the db and link it to the playlist */
         midb_add_record(midb, mi);
         p->files[p->num_files] = mi;
         warnx("in playlist \"%s\", file \"%s\" is NOT in meta DB (added for now)",
            p->name, media_file);
      }

      p->num_files++;
   }

   fclose(fin);
   return p;
}

/* Builds an array of all files in the given directory with a '.playlist'
 * extension, returning the number of such files found.
 *
 * Parameters:
 *    dirname        C-string of directory containing playlist files
 *
 *    filenames      Pointer to an array of C-strings that will be allocated
 *                   and built in this function.  This array is where the
 *                   filename of each playlist will be stored.
 * Returns:
 *    The number of files with a '.playlist' extension that were found.
 *
 * Notes:
 *    All allocation of the filenames array is handled here.  It is the
 *    responsibility of the caller to free()
 *       1. each element of that array
 *       2. the array itself.
 */
int playlist_get_playlist_filenames(const char *dirname, char ***fnames)
{
   char   *glob_pattern;
   int     fcount;
   glob_t  files;

   /* build the search pattern */
   if (asprintf(&glob_pattern, "%s/*.playlist", dirname) == -1)
      errx(1, "failed in building glob pattern");

   /* get the files */
   if (glob(glob_pattern, 0, NULL, &files) != 0 && errno != 0)
      err(1, "failed to glob playlists directory");

   /* allocate & copy each of the filenames found into the filenames array */
   if ((*fnames = calloc(files.gl_pathc, sizeof(char*))) == NULL)
      err(1, "failed to allocate playlist filenames array");

   for (fcount = 0; fcount < files.gl_pathc; fcount++)
   {  if (asprintf(&((*fnames)[fcount]), "%s", files.gl_pathv[fcount]) == -1)
         errx(1, "failed to allocate filename for playlist");
   }

   /* cleanup */
   globfree(&files);
   free(glob_pattern);

   return fcount;
}

/* Save a playlist to file.  The file used will be the one specified in the
 * "filename" member of the playlist object specified.
 *
 * Parameters:
 *    p        Const pointer to playlist object to save.
 *
 * Returns:
 *    Nothing.
 */
void playlist_save(const playlist *p)
{
   FILE *fout;

   if ((fout = fopen(p->filename, "w")) == NULL)
      err(1, "failed to open playlist \"%s\"", p->filename);

   /* write each song to file */
   int i;
   for (i = 0; i < p->num_files; i++)
   {  if (fprintf(fout, "%s\n", p->files[i]->file) == -1)
         err(1, "failed to record playlist \"%s\"", p->filename);
   }

   fclose(fout);
}

/* Deletes playlist from disk and destroy's the object, free()'ing all memory.
 *
 * Parameters:
 *    p        Pointer to playlist object to be deleted.
 *
 * Returns:
 *    Nothing.
 */
void playlist_delete(playlist *p)
{
   /* delete file if the playlist is stored in a file */
   if (p->filename != NULL && unlink(p->filename) != 0)
      err(1, "failed to delete playlist \"%s\"", p->filename);

   /* destroy/free() all memory */
   playlist_free(p);
}

/* Remove a song from a given playlist (free()'s all memory associated with
 * that song).
 *
 * Parameters:
 *    p        Pointer to playlist to remove the song from
 *    index    Index of the song to remove
 *
 * Returns:
 *    Nothing.
 */
void playlist_remove_file(playlist *p, int index)
{
   if (index < 0 || index > p->num_files)
      errx(1, "playlist remove file: index %d out of range", index);

   int i;
   for (i = index + 1; i < p->num_files; i++)
      p->files[i - 1] = p->files[i];

   p->num_files--;
}

