
#include "uinterface.h"

/*****************************************************************************
 * UI Create / Destroy Functions
 ****************************************************************************/

/* Initialize a uinterface object, initializes ncurses, creating all windows
 * objects , and sets up all attributes used.
 * The only variable needed to build the uinterface is the width of the
 * playlists window.
 *
 * Parameters:
 *    ui                Pointer to uinterface object to create.
 *    playlist_width    Width (in columns) of the playlists window on the left
 *                      of the screen.
 *
 * Returns:
 *    Nothing.
 *
 * Note:
 *    The uinterface object passed by pointer must already be allocated.
 *    All else is done here.  To destroy a uinterface, use ui_destroy.
 *
 *    TODO eventually, this needs to handle resizing of term.  This may be
 *       handled elsewhere (a call to ui_destroy and ui_init?), or not...
 *       first I have to figure out how to catch and handle resizes!
 */
void ui_init(uinterface *ui, int playlist_width)
{
   /* first, init curses */
   initscr();              /* init curses */
   cbreak();               /* disable buffering of input */
   noecho();               /* disable echo'ing of input */
   nonl();                 /* don't translate return keys into newline */
   keypad(stdscr, TRUE);   /* enable function keys, arrow keys, etc. */
   halfdelay(1);           /* XXX eeeeh... not sure if i like this */
   start_color();          /* allow colors */
   curs_set(0);            /* hide the cursor */
   ESCDELAY = 0;           /* eliminate 1 second delay after hitting 'ESC' */
   refresh();

   int lines, cols;
   getmaxyx(stdscr, lines, cols);
   ui->height     = lines;
   ui->width      = cols;
   ui->attr_error = 0;
   ui->attr_bars  = 0;

   /* TODO I know the above will have to be changed/reworked to allow for
    * resizing... but I still have no clue how to do this at all, let alone
    * cleanly.
    */

   /* allocate the four windows */
   ui->player    = calloc(1, sizeof(window));
   ui->command   = calloc(1, sizeof(window));
   ui->playlists = calloc(1, sizeof(window));
   ui->medialist = calloc(1, sizeof(window));
   if (ui->player    == NULL || ui->command   == NULL
   ||  ui->playlists == NULL || ui->medialist == NULL)
      err(1, "couldn't allocate user interface");

   /* initalize the 4 windows */
   /*             window         width            height      sx    sy */
   ui->playlist_width = playlist_width;
   ui_window_init(ui->player,    cols, 1, 0, 0);
   ui_window_init(ui->command,   cols, 1, 0, lines - 1);
   ui_window_init(ui->playlists, ui->playlist_width,            lines - 3,
      0, 2);
   ui_window_init(ui->medialist, cols - ui->playlist_width - 1, lines - 3,
      ui->playlist_width + 1, 2);

   ui->previous_window = ui->active_window = UI_PLAYLISTS;
}

/* Destroy's a given uinterface and free()'s all associated memory.
 *
 * Parameters:
 *    ui          Pointer to the uinterface object to destroy.
 *
 * Returns:
 *    Nothing
 *
 * Notes:
 *    All windows are destroyed and their memory free()'ed.
 *    Other settings of the ui are set to sane 'empty' defaults.
 *    The uinterface object given is NOT free()'ed.
 */
void ui_destroy(uinterface *ui)
{
   /* destroy each window (this also free()'s the mem for each window) */
   ui_window_destroy(ui->player);
   ui_window_destroy(ui->playlists);
   ui_window_destroy(ui->medialist);
   ui_window_destroy(ui->command);

   /* reset other properties of the ui to sane 'empty' defaults */
   ui->height = 0;
   ui->width  = 0;
   ui->previous_window = UI_PLAYLISTS;
   ui->active_window = UI_PLAYLISTS;

   /* end ncurses */
   endwin();
}

void ui_resize(uinterface *ui)
{
   struct winsize ws;

   /* get new dimensions */
   if (ioctl(STDIN_FILENO, TIOCGWINSZ, &ws) < 0)
      err(1, "ioctl for new size failed");

   if (ws.ws_row == LINES && ws.ws_col == COLS)
      return;

   /* update main ui & ui struct's width/height */
   resizeterm(ws.ws_row, ws.ws_col);
   ui->width  = ws.ws_col;
   ui->height = ws.ws_row;

   /* resize player window */
   ui_window_resize(ui->player, ws.ws_col, 1);

   /* resize command window, and move it */
   ui_window_resize(ui->command, ws.ws_col, 1);
   mvwin(ui->command->cwin, ws.ws_row - 1, 0);

   /* resize playlist and medialist windows */
   ui_window_resize(ui->playlists, ui->playlist_width, ws.ws_row - 3);
   ui_window_resize(ui->medialist, ws.ws_col - ui->playlist_width - 1, ws.ws_row - 3);

   wclear(ui->player->cwin);
   wclear(ui->command->cwin);
   wclear(ui->medialist->cwin);
   wclear(ui->playlists->cwin);
   clear();

   /* redraw all windows */
   wrefresh(ui->player->cwin);
   wrefresh(ui->command->cwin);
   wrefresh(ui->playlists->cwin);
   wrefresh(ui->medialist->cwin);
   refresh();
}

/*****************************************************************************
 * Window Create / Destroy Functions
 ****************************************************************************/

/* Initializes a window with the given parameters.  Other properties of the
 * window are set to sane defaults.
 *
 * Parameters:
 *    w           Pointer to window to init
 *    width       Width of window (number of columns on the screen)
 *    height      Height of window (number of rows on the screen)
 *    startx      Column of upper-left corner of the window on the main screen
 *    starty      Row of upper-left corner of the window on the main screen
 *
 * Returns:
 *    Nothing
 *
 * Notes:
 *    The only memory allocated is for the "row_fmt" member.  That needs to be
 *    free()'ed when destroyed (use ui_window_destroy for that)
 */
void ui_window_init(window *w, int width, int height, int startx, int starty)
{
   /* set all properties that were passed */
   w->cwin   = newwin(height, width, starty, startx);
   w->width  = width;
   w->height = height;

   /* set the remaining properties to sane defaults */
   w->voffset  = 0;
   w->hoffset  = 0;

   w->cur_row  = 0;
   w->max_rows = 0;

   w->attr_default  = 0;
   w->attr_cur_row  = 0;
   w->attr_cur_play = 0;

   asprintf(&(w->row_fmt_r), "%%%d.%ds", width, width);
   asprintf(&(w->row_fmt_l), "%%-%d.%ds", width, width);
   if (w->row_fmt_r == NULL || w->row_fmt_l == NULL)
      err(1, "ui_window_init: failed to create row fmt's");
}

void ui_window_resize(window *w, int new_width, int new_height)
{
   wresize(w->cwin, new_height, new_width);

   w->width  = new_width;
   w->height = new_height;

   free(w->row_fmt_r);
   free(w->row_fmt_l);
   asprintf(&(w->row_fmt_r), "%%%d.%ds", new_width, new_width);
   asprintf(&(w->row_fmt_l), "%%-%d.%ds", new_width, new_width);
   if (w->row_fmt_r == NULL || w->row_fmt_l == NULL)
      err(1, "ui_window_init: failed to create row fmt's");

   /* TODO this fixes cur-row if window shrinks vertically, but should also
    * adjust voffset to keep view "similar" ... ?
    */
   if (w->cur_row > new_height)
      w->cur_row = new_height - 1;
}

/* Destroys a window and free()'s all associated memory, including the window
 * itself.
 *
 * Parameters:
 *    w        Pointer to window to destroy
 *
 * Returns:
 *    Nothing
 */
void ui_window_destroy(window *w)
{
   /* free the ncurses window and all other memory, including window itself */
   delwin(w->cwin);
   free(w);
}

/*****************************************************************************
 * UI-Wide Functions
 ****************************************************************************/

/* Switch the focus'ed window of a uinterface to the window given.  Note that
 * this has no visual implications, so no window is redrawn.  This property
 * is simply used to keep track of which window input should go to (which is
 * done in another module).
 * Really, it's just for record keeping.
 *
 * This could easily be extended to allow some visual indicatio of what window
 * currently has focus, but as of right now, I don't think that's necessary.
 *
 * Parameters:
 *    ui          Pointer to uinterface object to work with
 *    new_win     WindowFocus (enum) of which of the 4 windows to assign
 *                focus to.
 *
 * Returns:
 *    Nothing.
 */
void ui_set_focus(uinterface *ui, WindowFocus new_win)
{
   ui->previous_window = ui->active_window;
   ui->active_window = new_win;
}

/* Draw the horizontal and vertical borders to the screen.  The horizontal
 * line is used to seperate the player window from those below, and the
 * vertical border is used to seperate the playlists window from the medialist
 * window.
 * 
 * Parameters:
 *    ui       Pointer to the uinterface object to work with
 *
 * Returns:
 *    Nothing.
 *
 * Notes:
 *    A refresh of the stdscr is done, as the lines are all drawn there.
 */
void ui_draw_borders(uinterface *ui)
{
   attron(ui->attr_bars);
   mvhline(1, 0, ACS_HLINE, ui->width);
   mvvline(1, ui->playlist_width, ACS_VLINE, ui->height - 2);
   mvaddch(1, ui->playlist_width, ACS_TTEE);
   attroff(ui->attr_bars);

   refresh();
}


/*****************************************************************************
 * General Window Functions
 ****************************************************************************/

void ui_window_change_row(window *w, int diff)
{
   /* TODO: this doesn't actually use diff yet */
   if (diff == 0)
      return;

   if (diff < 0)  /* move UP */
   {
      if (w->cur_row == 0)
      {  if (w->voffset > 0)
            w->voffset--;
      } else
         w->cur_row--;

   } else         /* move DOWN */
   {
      if (w->cur_row == w->height - 1)
      {  if (w->voffset + w->height < w->max_rows)
            w->voffset++;
      } else if (w->cur_row < w->max_rows - 1)
         w->cur_row++;

   }
}

void ui_window_vscroll(window *w, int diff)
{
   w->voffset += diff;
   if (w->voffset < 0)
      w->voffset = 0;
   else if (w->voffset > w->max_rows - w->height)
   {  w->voffset = w->max_rows - w->height;
      if (w->voffset < 0)
         w->voffset = 0;
   }
}

void ui_window_hscroll(window *w, int diff)
{
   w->hoffset += diff;

   if (w->hoffset < 0)
      w->hoffset = 0;
}

/*****************************************************************************
 * Functions for each Specific Window
 ****************************************************************************/

int ui_command_get_string(uinterface *ui, const char *prompt, char **dest)
{
   /* display the prompt */
   wattron(ui->command->cwin, ui->command->attr_default);
   mvwprintw(ui->command->cwin, 0, 0, "%s", prompt);
   wattroff(ui->command->cwin, ui->command->attr_default);

   /* position the cursor */
   curs_set(2);
   wmove(ui->command->cwin, 0, strlen(prompt));
   wrefresh(ui->command->cwin);

   /* start getting input */
   char *input = calloc(1000, sizeof(char));

   int  pos, ch;
   pos = 0;
   while ((ch = getch()))
   {
      if (ch == ERR)
         continue;

      if (ch == '\n' || ch == 13)
         break;

      if (ch == 27)  /* user hit 'ESC' - exit immediately */
      {  free(input);
         return UI_USER_CANCELLED;
      }

      /* FIXME my backspace key sends a '127', which matches none of the
       * curses KEY_* constants... wtf?
       */
      if (ch == 127    || ch == KEY_BACKSPACE || ch == KEY_LEFT
       || ch == KEY_DC || ch == KEY_SDC)
      {  if (pos == 0)
            beep();
         else
         {  mvwaddch(ui->command->cwin, 0, strlen(prompt) + pos - 1, ' ');
            wrefresh(ui->command->cwin);
            pos--;
         }
         continue;
      }

      input[pos] = ch;
      mvwaddch(ui->command->cwin, 0, strlen(prompt) + pos, ch);
      wrefresh(ui->command->cwin);
      pos++;
   }
   input[pos] = '\0';

   /* trim the fat */
   *dest = calloc(strlen(input) + 1, sizeof(char));
   snprintf(*dest, strlen(input) + 1, "%s", input);
   free(input);

   curs_set(0);

   return UI_OK;
}

int ui_command_get_yesno(uinterface *ui, const char *msg, int *response)
{
   char *answer;
   if (ui_command_get_string(ui, msg, &answer) == UI_USER_CANCELLED)
      return UI_USER_CANCELLED;

   if (strncasecmp(answer, "yes", 3) == 0
   ||  strncasecmp(answer, "y", 1) == 0)
      *response = 1;
   else if (strncasecmp(answer, "no", 2) == 0
         || strncasecmp(answer, "n", 1) == 0)
      *response = 0;
   else
      *response = -1;

   free(answer);
   return UI_OK;
}

int ui_get_number(uinterface *ui, int *num, char first)
{
   *num = 0;
   if (isdigit(first))
      *num = *num * 10 + first - 48;
   else
      return UI_USER_CANCELLED;

   int ch;
   while ((ch = getch()))
   {
      if (ch == ERR)
         continue;

      if (ch == 27)  /* user hit ESCAPE */
         return UI_USER_CANCELLED;

      if (isdigit(ch))
         *num = *num * 10 + ch - 48;
      else
         break;
   }

   return UI_OK;
}
