
#include <sys/time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <pwd.h>

#include "vitunes.h"

#include "viui.h"
#include "input_handler.h"


/* FIXME this is just a quick test of sorting */
int
meta_compare(const void *A, const void *B)
{
   const meta_info **a2 = (const meta_info**) A;
   const meta_info **b2 = (const meta_info**) B;

   const meta_info *a = *a2;
   const meta_info *b = *b2;

   if (a == NULL || b == NULL)
      errx(1, "in sorting playlist, two NULL's were encountered");

   int search_order[] = { MI_CINFO_ARTIST, MI_CINFO_ALBUM, MI_CINFO_TITLE,
                          MI_CINFO_TRACK };
   int search_order_size = 4;

   int i;
   for (i = 0; i < search_order_size; i++)
   {
      int field = search_order[i];
      if (a->cinfo[field] == NULL && b->cinfo[field] == NULL)
         return 0;
      else if (a->cinfo[field] != NULL && b->cinfo[field] == NULL)
         return -1;
      else if (a->cinfo[field] == NULL && b->cinfo[field] != NULL)
         return 1;
      else
      {  int ret = strcasecmp(a->cinfo[field], b->cinfo[field]);
         if (ret != 0)
            return ret;
      }
   }

   return 0;
}


/* fixed locations of various things (the %s in all is the user's homedir) */
const char *VITUNES_DIR_FMT  = "%s/.vitunes";
const char *CONFIG_FILE_FMT  = "%s/.vitunes/vitunes.conf";
const char *DB_FILE_FMT      = "%s/.vitunes/vitunes.db";
const char *PLAYLIST_DIR_FMT = "%s/.vitunes/playlists";

/* global vars corresponding to the above */
char *VITUNES_DIR;
char *CONFIG_FILE;
char *DB_FILE;
char *PLAYLIST_DIR;


/* global var to track when/if to quit vitunes */
bool  QUIT_VITUNES;

/* global signal flags and their signal handlers */
volatile sig_atomic_t gotwinch = 0;
void sig_winch(int);


/* timer to update media player. this is what handles stuff like keeping track
 * of where we're at in a currently playing file, and starts the 'next' file
 * when on finishes.
 */
media_player *global_mp = NULL;
void timer_handler(int);


/* other local funcs */
int process_command_line(int argc, char *argv[]);


int
main(int argc, char *argv[])
{
   int   playlist_width;

   /* -----------------------------------------------------------------------
    * load config file and parse command line switches
    * -------------------------------------------------------------------- */

   /* get home dir and build locations of various files */
   struct passwd  *pwd;
   pwd = getpwuid(getuid());
   if (pwd == NULL)
      errx(1, "invalid user %d", getuid());

   asprintf(&VITUNES_DIR,   VITUNES_DIR_FMT,  pwd->pw_dir);
   asprintf(&CONFIG_FILE,   CONFIG_FILE_FMT,  pwd->pw_dir);
   asprintf(&DB_FILE,       DB_FILE_FMT,      pwd->pw_dir);
   asprintf(&PLAYLIST_DIR,  PLAYLIST_DIR_FMT, pwd->pw_dir);
   playlist_width = 25;

   /* TODO: actually load a config file here... */

   process_command_line(argc, argv);


   /* -----------------------------------------------------------------------
    * starting vitunes as normal... init stuff
    * -------------------------------------------------------------------- */

   /* create user interface and media database objects */
   media_player    mp;
   media_library  *ml;
   uinterface      ui;

   /* initialize/load media library (includes db and playlists) */
   ml = ml_new();
   ml_load(ml, DB_FILE, PLAYLIST_DIR);
   /* TODO this works... now make the sort order customizable */
   qsort(ml->playlists[0]->files, ml->playlists[0]->num_files,
      sizeof(meta_info*), meta_compare);

   /* initialize media player and setup its timer */
   media_player_init(&mp);

   /* setup signal handler for timer/alarm */
   global_mp = &mp;
   struct sigaction sig_act;
   if (sigemptyset(&sig_act.sa_mask) < 0)
      err(1, "failed to get empty signal set");

   sig_act.sa_flags = 0;
   sig_act.sa_handler = timer_handler;
   if (sigaction(SIGALRM, &sig_act, NULL) < 0)
      err(1, "failed to reigster signal handler");

   /* setup timer */
   struct itimerval timer;
   timer.it_value.tv_sec = 0;
   timer.it_value.tv_usec = 500000;
   timer.it_interval.tv_sec = 0;
   timer.it_interval.tv_usec = 500000;
   if (setitimer(ITIMER_REAL, &timer, NULL) < 0)
      err(1, "failed to set timer");

   /* initalize user interface and signal handler for resize */
   ui_init(&ui, playlist_width);
   signal(SIGWINCH, sig_winch);

   /* load some defaults */
   /* TODO move this to a load-defaults function */
   init_pair(1, COLOR_RED,     COLOR_BLACK);
   init_pair(2, COLOR_CYAN,    COLOR_BLACK);
   init_pair(3, COLOR_MAGENTA, COLOR_BLACK);
   init_pair(4, COLOR_GREEN,   COLOR_BLACK);
   init_pair(5, COLOR_YELLOW,  COLOR_BLACK);

   ui.attr_error = COLOR_PAIR(1) | WA_BOLD;
   ui.attr_bars  = WA_DIM;
   ui.player->attr_default  = COLOR_PAIR(5);
   ui.command->attr_default = COLOR_PAIR(2);
   ui.playlists->attr_default  = COLOR_PAIR(3);
   ui.playlists->attr_cur_row  = WA_REVERSE;
   ui.playlists->attr_cur_play = WA_BOLD;
   ui.medialist->attr_default  = 0;
   ui.medialist->attr_cur_row  = WA_REVERSE;
   ui.medialist->attr_cur_play = COLOR_PAIR(4) | WA_BOLD;

   /* build vitunes struct */
   vitunes v;
   v.ui      = &ui;
   v.library = ml;
   v.mplayer = &mp;
   v.viewing_playlist = ML_PINDEX_LIBRARY;
   v.playing_playlist = -1;
   v.copy_buff_size   = 0;

   /* draw initial display */
   ui.playlists->max_rows  = ml->playlists_num;
   ui.medialist->max_rows  = ml->playlists[ML_PINDEX_LIBRARY]->num_files;
   ui_draw_borders(&ui);
   viui_draw_playlists(&v);
   viui_draw_medialist(&v);


   /* -----------------------------------------------------------------------
    * begin input loop
    * -------------------------------------------------------------------- */

   int input;
   QUIT_VITUNES = false;
   while ((input = wgetch(stdscr)) && !QUIT_VITUNES)
   {
      /* update player window */
      viui_draw_player(&v);

      /* handle any resize event */
      if (gotwinch)
      {  gotwinch = 0;
         ui_resize(v.ui);
         ui_draw_borders(v.ui);
         viui_draw_playlists(&v);
         viui_draw_medialist(&v);
         viui_draw_player(&v);
      }

      /* have input, will handle... */
      if (input != ERR)
         handle_input(&v, input);
   }


   /* -----------------------------------------------------------------------
    * exit vitunes / clenaup
    * -------------------------------------------------------------------- */

   /* ignore all registered signals */
   signal(SIGALRM,  SIG_IGN);
   signal(SIGWINCH, SIG_IGN);

   /* free all vitunes components */
   ui_destroy(&ui);
   ml_free(ml);
   media_player_destroy(&mp);
   if (v.copy_buff_size > 0)
      free(v.copy_buff);

   return 0;
}

void
usage(const char *pname)
{
   fprintf(stderr,"\
usage: %s [-h] [-e COMMAND ...]\n\
See \"%s -e help\" for information about what e-commands are available.\n\
",
   pname, pname);
   exit(1);
}

void
sig_winch(int dummy)
{
   gotwinch = 1;
}

void
timer_handler(int sig)
{
   media_player_monitor(global_mp);
}


/*****************************************************************************
 *
 *                            E - Commands
 *
 * What follows are all of the "E-Command" functions... that is, the functions
 * that are executed when "vitunes -e commandname" is entered.
 * The name of each function is preceeded with "ecmd_", and each takes the
 * form of a argc+argv[] argument list.
 *
 ****************************************************************************/

int
ecmd_createdb(int argc, char *argv[])
{
   if (argc <= 1)
      errx(1, "usage: -e %s /path/to/fileDir1 [ ... ]", argv[0]);

   /* create vitunes directory */
   if (mkdir(VITUNES_DIR, S_IRWXU) == -1)
   {  if (errno == EEXIST)
         warnx("vitunes directory \"%s\" already exists (OK)", VITUNES_DIR);
      else
         err(1, "unable to create vitunes directory '%s'", VITUNES_DIR);
   } else
      warnx("vitunes directory \"%s\" created", VITUNES_DIR);

   /* create playlists directory */
   if (mkdir(PLAYLIST_DIR, S_IRWXU) == -1)
   {  if (errno == EEXIST)
         warnx("playlists directory \"%s\" already exists (OK)", PLAYLIST_DIR);
      else
         err(1, "unable to create playlists directory '%s'", PLAYLIST_DIR);
   } else
      warnx("playlists directory \"%s\" created", PLAYLIST_DIR);

   /* check if db already exists */ 
   struct stat sb;
   if (stat(DB_FILE, &sb) < 0)
   {  if (errno != ENOENT)
         err(1, "failed to stat \"%s\"", DB_FILE);
   } else
   {  printf("WARNING: A database file already exists at \"%s\".\n", DB_FILE);
      printf("\tContinuing will erase that database entirely.\n");
      printf("\tIf you already have a working database, you should run the\n");
      printf("\tcommand \"updatedb\" instead.\n");
      printf("Do you wish to continue? [y/n] ");

      char  *response;
      size_t res_size;
      if ((response = fgetln(stdin, &res_size)) == NULL)
         err(1, "fgetln");

      response[res_size] = '\0';
      if (strcasecmp(response, "yes") != 0 && strcasecmp(response, "y") != 0)
      {  printf("creatdb cancelled.\n");
         return 1;
      }
   }

   midb_create(DB_FILE, argv + 1);
   return 0;
}

int
ecmd_updatedb(int argc, char *argv[])
{
   printf("Updating existing database...\n");
   midb_update(DB_FILE);

   /* if additional files/dirs were specified on the command line, load them */
   if (argc > 1)
   {  printf("\n\nAdding new files...\n");
      midb_create(DB_FILE, argv + 1);
   }

   return 0;
}

int
ecmd_flushdb(int argc, char *argv[])
{
   if (argc != 1)
      errx(1, "usage: -e %s", argv[0]);

   meta_info_db *midb = midb_new();
   midb_load(DB_FILE, midb);
   midb_flush(midb, stdout);
   midb_free(midb);

   return 0;
}

void
ecmd_help_createdb(void)
{  printf("\
VITUNES COMMAND:\n\tcreatedb\n\
SYNOPSIS:\n\tcreatedb /path/to/music [ /path/to/more/music ... ]\n\
DESCRIPTION:\n\
   The createdb command is used to build the initial meta information\n\
   database used by vitunes.  It scans all files and directories specified\n\
   on the command line for media files.  Each file found is scanned, and if\n\
   the file contains any meta information, it is added to the newly\n\
   created database.\n\n\
   Directories are searched recursively.\n\n\
   The createdb command will over-write any existing meta information\n\
   database, but you will be prompted should this be the case.\n\n\
");
}

void
ecmd_help_updatedb(void)
{  printf("\
VITUNES COMMAND:\n\tupdatedb\n\
SYNOPSIS:\n\tupdatedb [ fileDir1 fileDir2 ... ]\n\
DESCRIPTION:\n\
   The updatedb command loads the existing meta information database used\n\
   by vitunes and for each media file listed in that database, the file is\n\
   checked to see if it has been removed or modified since it was added\n\
   to the database.\n\n\
   If the file has been removed, it will be removed from the database.\n\
   If the file has been modified, it's meta information will be extracted\n\
   again and the database will be updated.\n\n\
   Additionally, one can specify files/directories that will be searched\n\
   (directories are searched recursively) for any media files, and if found,\n\
   they will be scanned and added to the database.\n\n\
   In short, anytime you remove/modify media files already in the vitunes\n\
   database, you should run\n\n\
         vitunes -e updatedb\n\n\
   If you want to add new music to database, run\n\n\
         vitunes -e updatedb /path/to/new/music1 /path/to/new/music2 ...\n\n\
");
}

void
ecmd_help_flushdb(void)
{  printf("\
VITUNES COMMAND:\n\tflushdb\n\
SYNOPSIS:\n\tflushdb\n\
DESCRIPTION:\n\
   The flushdb command simply outputs the contents of the meta information\n\
   database used by vitunes to stdout, in a fairly easy to read (but very\n\
   to parse/grep through) formst.\n\n\
   This command is not intended to be used directly for any real purpose.\n\
   It is only provided for debugging purposes.\n\
");
}

int
ecmd_help(int argc, char *argv[])
{
   if (argc > 2)
      errx(1, "usage: -e %s [command]", argv[0]);

   /* no help requested for a specific command, give a list of all e-cmds */
   if (argc == 1)
   {  printf("\
The following is a list of \"e-commands\" that you execute by doing\n\n\
                        vitunes -e COMMAND [args ...]\n\n\
All of the commands work with the vitunes meta-information database in one\n\
way or another.\n\
For each command below, use the following to get detailed information about\n\
the usage of each command:\n\
                        vitunes -e help COMMAND\n\n\
   COMMAND     DESCRIPTION\n\
   -------     ------------------------------------------------------------\n\
   createdb    Create the initial meta-information database and general\n\
               vitunes structure in \"~/.vitunes/\".  A list of files and/or\n\
               directories are provided on the command line and scanned for\n\
               media files containing meta-information (directories are\n\
               searched recursively).\n\
               This command should only be run when vitunes is first\n\
               installed, or after an upgrade.\n\n\
   updatedb    Load the existing database and check all media files\n\
               contained within it to see if any of they have been removed\n\
               or if their meta information has changed.  The library is\n\
               updated accordingly.\n\
               Additionally, files/directories can be specified on that will\n\
               be searched for new files to add.\n\
               This command should be run whenever you update your media\n\
               files (change their info, add new ones, removed files, etc.)\n\n\
   flushdb     Load the existing database and dump it's information in an\n\
               easy-to-parse format to stdout.\n\n\
   help        This command.\n\
");
      return 0;
   }

   /* if reach here, help fora specific command was requested */
   if (strcmp(argv[1], "createdb") == 0)
      ecmd_help_createdb();
   else if (strcmp(argv[1], "updatedb") == 0)
      ecmd_help_updatedb();
   else if (strcmp(argv[1], "flushdb") == 0)
      ecmd_help_flushdb();
   else if (strcmp(argv[1], "help") == 0)
      printf("You're a damn fool if you need help with help.\n");
   else
      printf("Unknown command \"%s\".  See \"vitunes -e help\" for list.\n",
         argv[1]);

   return 0;
}

/* array of the above functions and their names */
struct ecmd_entry {
   char  *name;
   int   (*fname)(int argc, char *argv[]);
};

const struct ecmd_entry ECMD_PATH[] = {
   { "createdb",  ecmd_createdb },
   { "updatedb",  ecmd_updatedb },
   { "flushdb",   ecmd_flushdb },
   { "help",      ecmd_help }
};
const int ECMD_PATH_SIZE = (sizeof(ECMD_PATH) / sizeof(struct ecmd_entry));


/*****************************************************************************
 *                         The command line parser
 ****************************************************************************/
int
process_command_line(int argc, char *argv[])
{
   int ch;
   int i;

   while ((ch = getopt(argc, argv, "he:")) != -1)
   {
      switch (ch)
      {
         case 'e':
            argc -= optind - 1;
            argv += optind - 1;

            for (i = 0; i < ECMD_PATH_SIZE; i++)
            {  if (strcmp(optarg, ECMD_PATH[i].name) == 0)
                  exit((ECMD_PATH[i].fname)(argc, argv));
            }

            errx(1, "Unknown e-command '%s'.  See 'vitunes -e help' for list.",
               optarg);
            break;


         case 'h':
         case '?':
         default:
            usage(argv[0]);
            /* NOT REACHED */
      }
   }

   return 0;
}
